"""Shared utilities."""

from .api_utils import query_radarr, query_sonarr, query_tmdb
from .ffmpeg_utils import channels_to_layout, get_probe_as_box, run_ffprobe

from .cli import coerce_video_files, create_default_config, parse_trait_filters  # isort: skip

__all__ = [
    "channels_to_layout",
    "coerce_video_files",
    "create_default_config",
    "get_probe_as_box",
    "parse_trait_filters",
    "query_radarr",
    "query_sonarr",
    "query_tmdb",
    "run_ffprobe",
]



"""Inspect subcommand."""

import cappa
from nclutils import console

from vid_cleaner.utils import coerce_video_files
from vid_cleaner.vidcleaner import InspectCommand
from vid_cleaner.views import stream_table


def main(inspect_cmd: InspectCommand) -> None:
    """Inspect video files.

    Inspect video files and print detailed information about them.

    Args:
        cmd (VidCleaner): The main command instance containing global options
        inspect_cmd (InspectCommand): The inspect subcommand instance with inspect-specific options

    Raises:
        cappa.Exit: If the video files are not found or if the inspect command is not valid
    """
    for video in coerce_video_files(inspect_cmd.files):
        if inspect_cmd.json_output:
            console.print(video.ffprobe_json())
            continue

        console.print(stream_table(video.probe_box))

    raise cappa.Exit(code=0)



"""Search subcommand."""

import cappa
from nclutils import console, find_files, find_subdirectories
from rich.live import Live
from rich.progress import track
from rich.table import Table

from vid_cleaner.config import settings
from vid_cleaner.constants import VideoContainerTypes
from vid_cleaner.utils import coerce_video_files
from vid_cleaner.vidcleaner import SearchCommand


def main(search_cmd: SearchCommand) -> None:
    """Search for video files under a directory.

    Args:
        search_cmd (SearchCommand): The search command instance with search-specific options

    Raises:
        cappa.Exit: If no video files are found
    """
    human_readable_filters = ", ".join(f"'{f.value}'" for f in settings.filters)

    directories_to_search = (
        [*find_subdirectories(search_cmd.directory, depth=search_cmd.depth), search_cmd.directory]
        if search_cmd.depth > 0
        else [search_cmd.directory]
    )

    video_files = []

    with Live(console=console, auto_refresh=True) as live:
        for i, directory in enumerate(directories_to_search):
            video_files.extend(
                coerce_video_files(
                    find_files(
                        directory,
                        globs=[
                            f"*{container_type.value}" for container_type in VideoContainerTypes
                        ],
                    )
                )
            )
            live.update(
                f"Found {len(video_files)} video files in {i + 1}/{len(directories_to_search)} directories"
            )

        live.update(
            f"[dim]Found {len(video_files)} video files in {i + 1}/{len(directories_to_search)} directories[/dim]"
        )
        live.stop()

    if len(video_files) == 0:
        console.print(f"No video files found in {search_cmd.directory}")
        raise cappa.Exit(code=0)

    table = Table(title="Video Files")
    table.add_column("#")
    table.add_column("Name")
    table.add_column("Matching filters")
    table.add_column("Video info")

    i = 0
    for video_file in track(
        video_files,
        description=f"Filtering {len(video_files)} files for {human_readable_filters}...",
        transient=True,
    ):
        video_traits = video_file.get_traits()

        matches = [trait for trait in video_traits if trait in settings.filters]
        if settings.filters and not matches:
            continue

        i += 1
        table.add_row(str(i), video_file.name, ", ".join(matches), ", ".join(video_traits))

    if i == 0:
        console.print(f"No video files found matching {human_readable_filters}")
        raise cappa.Exit(code=1)

    console.print(table)

    raise cappa.Exit(code=0)



"""Instantiate settings default values."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import cappa
from dynaconf import Dynaconf, ValidationError, Validator
from nclutils import pp
from validators import url as url_validate

from vid_cleaner.constants import CACHE_DIR, DEFAULT_CONFIG_PATH, DEV_CONFIG_PATH, USER_CONFIG_PATH


@dataclass
class SettingsManager:
    """Manage application settings through a singleton pattern.

    Provide centralized configuration management with support for default values, project-specific overrides, and CLI argument integration. Handles initialization of settings from config files, environment variables, and runtime overrides while maintaining type safety through validators.
    """

    _instance: Dynaconf | None = None

    @classmethod
    def initialize(cls) -> Dynaconf:
        """Create and configure a new Dynaconf settings instance with default values.

        Configure settings with environment variables, config files, and validators for all supported settings. Return existing instance if already initialized.

        Returns:
            Dynaconf: The configured settings instance with all validators registered.

        Raises:
            cappa.Exit: If settings are not initialized or project name is not found in config.
        """
        if cls._instance is not None:
            return cls._instance

        settings = Dynaconf(
            environments=False,
            envvar_prefix="VIDCLEANER",
            settings_files=[DEFAULT_CONFIG_PATH, USER_CONFIG_PATH, DEV_CONFIG_PATH],
            validate_on_update="all",
        )

        # Register all validators at once
        settings.validators.register(
            Validator("cache_dir", default=CACHE_DIR),
            Validator(
                "keep_local_subtitles",
                default=False,
                cast=bool,
            ),
            Validator(
                "langs_to_keep",
                default=["en"],
                cast=list,
                condition=lambda x: isinstance(x, list),
                messages={"condition": "'{name}' must be a list"},
            ),
            Validator("overwrite", default=False, cast=bool),
            Validator(
                "keep_commentary",
                default=False,
                cast=bool,
            ),
            Validator(
                "drop_local_subs",
                default=False,
                cast=bool,
            ),
            Validator(
                "keep_all_subtitles",
                default=False,
                cast=bool,
            ),
            Validator(
                "drop_original_audio",
                default=False,
                cast=bool,
            ),
            Validator(
                "downmix_stereo",
                default=False,
                cast=bool,
            ),
            Validator("out_path", default=None),
            Validator(
                "radarr_url",
                default="",
                cast=str,
                condition=lambda x: url_validate(x) or not x,
                messages={"condition": "'{name}' in settings must be a valid URL: '{value}'"},
            ),
            Validator(
                "sonarr_url",
                default="",
                cast=str,
                condition=lambda x: url_validate(x) or not x,
                messages={"condition": "'{name}' in settings must be a valid URL: '{value}'"},
            ),
            Validator(
                "radarr_api_key",
                default="",
                cast=str,
                must_exist=True,
                condition=lambda x: x,
                when=Validator("radarr_url", must_exist=True, condition=lambda x: x),
                messages={"condition": "'{name}' must be set if 'radarr_url' is set"},
            ),
            Validator(
                "sonarr_api_key",
                default="",
                cast=str,
                must_exist=True,
                condition=lambda x: x,
                when=Validator("sonarr_url", must_exist=True, condition=lambda x: x),
                messages={"condition": "'{name}' must be set if 'sonarr_url' is set"},
            ),
        )

        try:
            settings.validators.validate_all()
        except ValidationError as e:
            accumulative_errors = e.details
            for error in accumulative_errors:
                pp.error(error[1])
            raise cappa.Exit(code=1) from e
        except ValueError as e:
            pp.error(str(e))
            raise cappa.Exit(code=1) from e

        cls._instance = settings
        return settings

    @classmethod
    def apply_cli_settings(cls, cli_settings: dict[str, Any]) -> None:
        """Override existing settings with non-None values from CLI arguments.

        Update the settings singleton with any non-None values provided via command line arguments, preserving existing values for unspecified settings.

        Args:
            cli_settings (dict[str, Any]): Dictionary of settings from CLI arguments to apply as overrides.

        Raises:
            cappa.Exit: If settings singleton has not been initialized.
        """
        settings = cls._instance
        if settings is None:  # pragma: no cover
            msg = "Settings not initialized"
            pp.error(msg)
            raise cappa.Exit(code=1)

        # Filter out None values to avoid overriding with None
        cli_overrides = {k: v for k, v in cli_settings.items() if v is not None}
        settings.update(cli_overrides)


# Initialize settings singleton
settings = SettingsManager.initialize()



"""Shared fixtures for tests."""

import json
from pathlib import Path

import pytest
from nclutils.pytest_fixtures import clean_stdout, debug  # noqa: F401
from rich.console import Console

from vid_cleaner.utils import get_probe_as_box

console = Console()


@pytest.fixture
def mock_video_path(tmp_path):
    """Fixture to return a VideoFile instance with a specified path.

    Returns:
        VideoFile: A VideoFile instance with a specified path.
    """
    # GIVEN a VideoFile instance with a specified path
    test_path = Path(tmp_path / "test_video.mp4")
    test_path.touch()  # Create a dummy file
    return test_path


@pytest.fixture
def mock_ffprobe_box(mocker):
    """Return mocked JSON response from ffprobe."""

    def _inner(filename: str):
        fixture = Path(__file__).resolve().parent / "fixtures/ffprobe" / filename

        cleaned_content = []  # Remove comments from JSON
        with fixture.open() as f:
            for line in f.readlines():
                # Remove comments
                if "//" in line:
                    continue
                cleaned_content.append(line)

        mocker.patch(
            "vid_cleaner.utils.ffmpeg_utils.run_ffprobe",
            return_value=json.loads("".join(line for line in cleaned_content)),
        )

        return get_probe_as_box(fixture)

    return _inner


@pytest.fixture
def mock_ffprobe():
    """Return mocked JSON response from ffprobe."""

    def _inner(filename: str):
        fixture = Path(__file__).resolve().parent / "fixtures/ffprobe" / filename

        cleaned_content = []  # Remove comments from JSON
        with fixture.open() as f:
            for line in f.readlines():
                # Remove comments
                if "//" in line:
                    continue
                cleaned_content.append(line)

        return json.loads("".join(line for line in cleaned_content))

    return _inner


@pytest.fixture
def mock_ffmpeg(mocker):
    """Fixture to mock the FfmpegProgress class to effectively mock the ffmpeg command and its progress output.

    Usage:
        def test_something(mock_ffmpeg):
            # Mock the FfmpegProgress class
            mock_ffmpeg_progress = mock_ffmpeg()

            # Test the functionality
            do_something()
            mock_ffmpeg.assert_called_once() # Confirm that the ffmpeg command was called once
            args, _ = mock_ffmpeg.call_args # Grab the ffmpeg command arguments
            command = " ".join(args[0]) # Join the arguments into a single string
            assert command == "ffmpeg -i input.mp4 output.mp4" # Check the command

    Returns:
        Mock: A mock object for the FfmpegProgress class.
    """
    mock_ffmpeg_progress = mocker.patch(
        "vid_cleaner.models.video_file.FfmpegProgress",
        autospec=True,
    )
    mock_instance = mock_ffmpeg_progress.return_value
    mock_instance.run_command_with_progress.return_value = iter([0, 25, 50, 75, 100])
    return mock_ffmpeg_progress



"""Utilities for CLI."""

import shutil
from pathlib import Path

import cappa
from nclutils import pp

from vid_cleaner.constants import (
    DEFAULT_CONFIG_PATH,
    USER_CONFIG_PATH,
    VideoContainerTypes,
    VideoTrait,
)
from vid_cleaner.models.video_file import VideoFile


def coerce_video_files(files: list[Path]) -> list[VideoFile]:
    """Parse and validate a list of video file paths.

    Verify each path exists and has a valid video container extension. Convert valid paths into VideoFile objects.

    Args:
        files (list[Path]): List of file paths to validate and convert

    Returns:
        list[VideoFile]: List of validated VideoFile objects

    Raises:
        cappa.Exit: If a file doesn't exist or has an invalid extension
    """
    for file in files:
        f = file.expanduser().resolve().absolute()

        if not f.is_file():
            msg = f"File '{file}' does not exist"
            raise cappa.Exit(msg, code=1)

        if f.suffix.lower() not in [container.value for container in VideoContainerTypes]:
            msg = f"File {file} is not a video file"
            raise cappa.Exit(msg, code=1)

    return [VideoFile(path.expanduser().resolve().absolute()) for path in files]


def create_default_config() -> None:
    """Create a default configuration file.

    Create a new configuration file at the default user location if one does not already exist. Copy the default configuration template to initialize the file.
    """
    if not USER_CONFIG_PATH.exists():
        USER_CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        USER_CONFIG_PATH.touch(exist_ok=True)
        shutil.copy(DEFAULT_CONFIG_PATH, USER_CONFIG_PATH)
        pp.info(f"Default configuration file created: `{USER_CONFIG_PATH}`")


def parse_trait_filters(facets: str) -> set[VideoTrait]:
    """Parse a comma-separated list of facets into a list of VideoTrait enums.

    Args:
        facets (str): Comma-separated string of facet names to parse

    Returns:
        set[SearchFacet]: Set of VideoTrait enum values

    Raises:
        cappa.Exit: If any facet is invalid, exits with code 1
    """
    try:
        return {VideoTrait(facet.lower()) for facet in facets.split(",")}
    except ValueError as e:
        pp.error(f"Invalid facet: {e}")
        raise cappa.Exit(code=1) from e



"""Helper utilities."""

from pathlib import Path

import cappa
import ffmpeg as python_ffmpeg
from box import Box
from nclutils import pp

from vid_cleaner.constants import AudioLayout, CodecTypes


def channels_to_layout(channels: int) -> AudioLayout | None:
    """Convert number of audio channels to an AudioLayout enum value.

    Convert a raw channel count into the appropriate AudioLayout enum value for use in audio processing. Handle special cases where 5 channels maps to SURROUND5 (5.1) and 7 channels maps to SURROUND7 (7.1).

    Args:
        channels (int): Number of audio channels in the stream

    Returns:
        AudioLayout | None: The corresponding AudioLayout enum value if a valid mapping exists,
            None if no valid mapping is found

    Examples:
        >>> channels_to_layout(2)
        <AudioLayout.STEREO: 2>
        >>> channels_to_layout(5)
        <AudioLayout.SURROUND5: 6>
        >>> channels_to_layout(7)
        <AudioLayout.SURROUND7: 8>
        >>> channels_to_layout(3)
    """
    if channels in [layout.value for layout in AudioLayout]:
        return AudioLayout(channels)

    if channels == 5:  # noqa: PLR2004
        return AudioLayout.SURROUND5

    if channels == 7:  # noqa: PLR2004
        return AudioLayout.SURROUND7

    return None


def run_ffprobe(path: Path) -> dict:  # pragma: no cover
    """Probe video file and return a dict.

    Args:
        path (Path): Path to video file

    Returns:
        dict: A dictionary containing information about the video file.

    Raises:
        cappa.Exit: If an error occurs while probing the video file.
    """
    try:
        probe = python_ffmpeg.probe(path)
    except python_ffmpeg.Error as e:
        pp.error(e.stderr)
        raise cappa.Exit(code=1) from e

    return probe


def get_probe_as_box(input_path: Path) -> Box:
    """Parse ffprobe output into a Box object with normalized stream and format data.

    Convert the raw ffprobe JSON output into a Box object with standardized fields. Extract and normalize metadata like title, format details, duration, and stream properties. Convert codec types and audio channels to enums.

    Args:
        input_path (Path): Path to the video file to probe

    Returns:
        Box: Box object containing normalized probe data with fields for format metadata and stream properties
    """
    probe_box = Box(
        run_ffprobe(input_path),
        default_box=True,
        default_box_create_on_get=False,
    )

    probe_box.path_to_file = input_path
    probe_box.name = probe_box.format.tags.title or probe_box.format.filename or input_path.name
    probe_box.format_name = probe_box.format.format_name or None
    probe_box.format_long_name = probe_box.format.format_long_name or None
    probe_box.duration = probe_box.format.duration or None
    probe_box.start_time = probe_box.format.start_time or None
    probe_box.size = probe_box.format.size or None
    probe_box.bit_rate = probe_box.format.bit_rate or None

    # Set stream codecs to enum
    for stream in probe_box.streams:
        stream.codec_type = CodecTypes(stream.codec_type.lower())
        stream.bps = stream.tags.BPS or None
        stream.title = stream.tags.title or None
        stream.channels = channels_to_layout(stream.channels)
        stream.language = stream.language or stream.tags.language or None

    return probe_box



"""Duty tasks for the project."""

from __future__ import annotations

import os
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from duty import duty, tools
from nclutils import console

if TYPE_CHECKING:
    from duty.context import Context

PY_SRC_PATHS = (Path(_) for _ in ("src/", "tests/", "duties.py", "scripts/") if Path(_).exists())
PY_SRC_LIST = tuple(str(_) for _ in PY_SRC_PATHS)
CI = os.environ.get("CI", "0") in {"1", "true", "yes", ""}
DEV_DIR = Path(__file__).parent.absolute() / ".development"


def strip_ansi(text: str) -> str:
    """Remove ANSI escape sequences from a string.

    Args:
        text (str): String to remove ANSI escape sequences from.

    Returns:
        str: String without ANSI escape sequences.
    """
    ansi_chars = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")

    # Replace [ with \[ so rich doesn't interpret output as style tags
    return ansi_chars.sub("", text).replace("[", r"\[")


def pyprefix(title: str) -> str:
    """Add a prefix to the title if CI is true.

    Returns:
        str: Title with prefix if CI is true.
    """
    if CI:
        prefix = f"(python{sys.version_info.major}.{sys.version_info.minor})"
        return f"{prefix:14}{title}"
    return title


@duty(silent=True)
def clean(ctx: Context) -> None:
    """Clean the project."""
    ctx.run("rm -rf .cache")
    ctx.run("rm -rf build")
    ctx.run("rm -rf dist")
    ctx.run("rm -rf pip-wheel-metadata")
    ctx.run("find . -type d -name __pycache__ | xargs rm -rf")
    ctx.run("find . -name '.DS_Store' -delete")


@duty
def ruff(ctx: Context) -> None:
    """Check the code quality with ruff."""
    ctx.run(
        tools.ruff.check(*PY_SRC_LIST, fix=False, config="pyproject.toml"),
        title=pyprefix("code quality check"),
        command="ruff check --config pyproject.toml --no-fix src/",
    )


@duty
def format(ctx: Context) -> None:  # noqa: A001
    """Format the code with ruff."""
    ctx.run(
        tools.ruff.format(*PY_SRC_LIST, check=True, config="pyproject.toml"),
        title=pyprefix("code formatting"),
        command="ruff format --check --config pyproject.toml src/",
    )


@duty
def mypy(ctx: Context) -> None:
    """Check the code with mypy."""
    os.environ["FORCE_COLOR"] = "1"
    ctx.run(
        tools.mypy("src/", config_file="pyproject.toml"),
        title=pyprefix("mypy check"),
        command="mypy --config-file pyproject.toml src/",
    )


@duty
def typos(ctx: Context) -> None:
    """Check the code with typos."""
    ctx.run(
        ["typos", "--config", ".typos.toml"],
        title=pyprefix("typos check"),
        command="typos --config .typos.toml",
    )


@duty(skip_if=CI, skip_reason="skip pre-commit in CI environments")
def precommit(ctx: Context) -> None:
    """Run pre-commit hooks."""
    ctx.run(
        "SKIP=mypy,pytest,ruff pre-commit run --all-files",
        title=pyprefix("pre-commit hooks"),
    )


@duty(pre=[ruff, mypy, typos, precommit], capture=CI)
def lint(ctx: Context) -> None:
    """Run all linting duties."""


@duty(capture=CI)
def update(ctx: Context) -> None:
    """Update the project."""
    ctx.run(["uv", "lock", "--upgrade"], title="update uv lock")
    ctx.run(["pre-commit", "autoupdate"], title="pre-commit autoupdate")


@duty()
def test(ctx: Context, *cli_args: str) -> None:
    """Test package and generate coverage reports."""
    ctx.run(
        tools.pytest(
            "tests",
            config_file="pyproject.toml",
            color="yes",
        ).add_args(
            "--cov",
            "--cov-config=pyproject.toml",
            "--cov-report=xml",
            "--cov-report=term",
            *cli_args,
        ),
        title=pyprefix("Running tests"),
        capture=CI,
    )


@duty()
def dev_clean(ctx: Context) -> None:
    """Clean the development environment."""
    # We import these here to avoid importing code before pytest-cov is initialized

    if DEV_DIR.exists():
        ctx.run(["rm", "-rf", str(DEV_DIR)])


@duty(pre=[dev_clean])
def dev_setup(ctx: Context) -> None:
    """Provision a mock development environment."""
    DEV_DIR.mkdir(parents=True, exist_ok=True)
    _create_subtitles()
    ctx.run(
        [
            "ffmpeg",
            "-f",
            "lavfi",
            "-i",
            "color=c=red:s=1920x1080:d=5:r=30",
            "-f",
            "lavfi",
            "-i",
            "sine=frequency=440:duration=5:sample_rate=48000",
            "-c:v",
            "libx264",
            "-vf",
            "format=yuv420p",
            "-preset",
            "medium",
            "-crf",
            "23",
            "-c:a",
            "aac",
            "-b:a",
            "128k",
            "-shortest",
            f"{DEV_DIR}/output_red_with_tone_1080p.mp4",
        ]
    )

    ctx.run(
        [
            "ffmpeg",
            "-f",
            "lavfi",
            "-i",
            "color=c=blue:s=1920x1080:d=5:r=30",
            "-f",
            "lavfi",
            "-i",
            "anullsrc=channel_layout=stereo:sample_rate=48000:d=5",
            "-f",
            "lavfi",
            "-i",
            "anullsrc=channel_layout=stereo:sample_rate=48000:d=5",
            "-f",
            "lavfi",
            "-i",
            "anullsrc=channel_layout=stereo:sample_rate=48000:d=5",
            "-i",
            f"{DEV_DIR}/subtitles/english.srt",
            "-i",
            f"{DEV_DIR}/subtitles/dutch.srt",
            "-i",
            f"{DEV_DIR}/subtitles/spanish.srt",
            "-map",
            "0:v:0",
            "-map",
            "1:a:0",
            "-metadata:s:a:0",
            "language=eng",
            "-metadata:s:a:0",
            "title=English Audio",
            "-map",
            "2:a:0",
            "-metadata:s:a:1",
            "language=dut",
            "-metadata:s:a:1",
            "title=Dutch Audio",
            "-map",
            "3:a:0",
            "-metadata:s:a:2",
            "language=spa",
            "-metadata:s:a:2",
            "title=Spanish Audio",
            "-map",
            "4:s:0",
            "-metadata:s:s:0",
            "language=eng",
            "-metadata:s:s:0",
            "title=English Subtitles",
            "-map",
            "5:s:0",
            "-metadata:s:s:1",
            "language=dut",
            "-metadata:s:s:1",
            "title=Dutch Subtitles",
            "-map",
            "6:s:0",
            "-metadata:s:s:2",
            "language=spa",
            "-metadata:s:s:2",
            "title=Spanish Subtitles",
            "-c:v",
            "libx264",
            "-vf",
            "format=yuv420p",
            "-preset",
            "medium",
            "-crf",
            "23",
            "-c:a",
            "aac",
            "-b:a",
            "128k",
            "-c:s",
            "copy",
            "-shortest",
            f"{DEV_DIR}/output_multilingual_1080p.mkv",
        ]
    )


def _create_subtitles() -> None:
    """Create subtitles for the development environment."""

    def _create_srt_file(filename, line1_text, line2_text) -> None:
        """Creates an SRT file with two subtitle entries.

        Args:
            filename (str): The name of the SRT file to create (e.g., "english.srt").
            lang_code (str): A short code for the language (e.g., "en" for English).
                            This is just for internal reference in this script,
                            the actual language metadata is set in ffmpeg.
            line1_text (str): The text for the first subtitle entry.
            line2_text (str): The text for the second subtitle entry.
        """
        content = f"""1
00:00:01,000 --> 00:00:04,000
{line1_text}

2
00:00:04,500 --> 00:00:05,000
{line2_text}
    """
        subtitle_folder = DEV_DIR / "subtitles"
        subtitle_folder.mkdir(parents=True, exist_ok=True)
        filename = subtitle_folder / filename
        try:
            with filename.open("w", encoding="utf-8") as f:
                f.write(content)
            console.print(f"Successfully created '{filename}'")
        except OSError as e:
            console.print(f"Error creating file '{filename}': {e}")

    subtitles_data = [
        {
            "filename": "english.srt",
            "line1": "This is an English subtitle.",
            "line2": "End.",
        },
        {
            "filename": "dutch.srt",
            "line1": "Dit is een Nederlandse ondertitel.",
            "line2": "Einde.",
        },
        {
            "filename": "spanish.srt",
            "line1": "Este es un subtítulo en español.",
            "line2": "Fin.",
        },
    ]

    for sub_info in subtitles_data:
        _create_srt_file(sub_info["filename"], sub_info["line1"], sub_info["line2"])



"""Test cache subcommand."""

import cappa
import pytest

from vid_cleaner import settings
from vid_cleaner.vidcleaner import VidCleaner, config_subcommand


@pytest.fixture
def tmp_cache_dir(tmp_path):
    """Create a temporary cache directory with test files."""
    # Given: A temporary cache directory
    cache_dir = tmp_path / "cache"
    cache_dir.mkdir()

    # And: A test file in the cache directory
    test_file = cache_dir / "test.txt"
    test_file.touch()

    # And: A directory with a test file
    test_dir = cache_dir / "test_dir"
    test_dir.mkdir()
    test_file = test_dir / "test.txt"
    test_file.touch()

    return cache_dir


def test_cache_list(tmp_cache_dir, clean_stdout, debug):
    """Verify cache list displays directory tree structure."""
    # Given: Cache directory path in settings
    args = ["cache"]
    settings.update({"cache_dir": tmp_cache_dir})

    # When: Invoking cache list command
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Output contains expected tree structure
    output = clean_stdout()
    assert exc_info.value.code == 0
    assert "│   └── 📄 " in output
    assert "├── 📂 " in output
    assert "└── 📄 " in output


def test_cache_clean(tmp_cache_dir, clean_stdout, debug):
    """Verify cache clean removes all cached files."""
    # Given: Cache directory path in settings
    args = ["cache", "-c"]
    settings.update({"cache_dir": tmp_cache_dir})

    # When: Invoking cache clean command
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Success message is displayed
    output = clean_stdout()
    assert exc_info.value.code == 0
    assert "Success: Cache cleared" in output

    # And: Cache directory is empty
    assert not any(tmp_cache_dir.iterdir())



"""Test the vidcleaner clip subcommand."""

from pathlib import Path

import cappa
import pytest

from vid_cleaner import settings
from vid_cleaner.vidcleaner import VidCleaner, config_subcommand


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        (["--start", "0:0"], "Error: --start must be in format HH:MM:SS"),
        (["--duration", "0:0"], "Error: --duration must be in format HH:MM:SS"),
    ],
)
def test_clip_option_errors(debug, tmp_path, clean_stdout, mock_video_path, args, expected):
    """Verify clip command validates time format arguments."""
    # Given: Invalid time format arguments
    args = ["clip", *args, str(mock_video_path)]
    settings.update({"cache_dir": Path(tmp_path), "langs_to_keep": ["en"]})

    # When: Running clip command with invalid arguments
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Error message is displayed
    output = clean_stdout()
    assert exc_info.value.code == 1
    assert expected in output


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        ([], "-ss 00:00:00 -t 00:01:00 -map 0"),
        (["--start", "00:05:00"], "-ss 00:05:00 -t 00:01:00 -map 0"),
        (["--start", "00:05:00", "--duration", "00:10:00"], "-ss 00:05:00 -t 00:10:00 -map 0"),
        (["--duration", "00:10:00"], "-ss 00:00:00 -t 00:10:00 -map 0"),
    ],
)
def test_clipping_video(
    mocker,
    mock_ffprobe_box,
    clean_stdout,
    mock_video_path,
    tmp_path,
    mock_ffmpeg,
    debug,
    args,
    expected,
):
    """Verify clip command extracts video segment with specified time range."""
    # Given: Mock video file and time range arguments
    args = ["clip", *args, str(mock_video_path)]
    settings.update({"cache_dir": Path(tmp_path), "langs_to_keep": ["en"]})

    # And: Mocked video metadata and output path

    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )
    mocker.patch("vid_cleaner.cli.clip_video.copy_file", return_value="clipped_video.mkv")

    # When: Running clip command
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # THEN verify ffmpeg was called with correct parameters
    mock_ffmpeg.assert_called_once()
    args, _ = mock_ffmpeg.call_args
    command = " ".join(args[0])
    assert expected in command

    # And: Success message is displayed
    assert exc_info.value.code == 0
    assert "✅ Success: clipped_video.mkv" in output


@pytest.mark.parametrize(
    ("args"),
    [
        ([]),
        (["--start", "00:05:00"]),
        (["--start", "00:05:00", "--duration", "00:10:00"]),
        (["--duration", "00:10:00"]),
    ],
)
def test_clipping_video_dryrun(
    mocker,
    clean_stdout,
    mock_ffprobe_box,
    mock_video_path,
    tmp_path,
    mock_ffmpeg,
    debug,
    args,
):
    """Verify clip command dry-run shows command without execution."""
    # Given: Mock video file and dry-run flag
    args = ["clip", "-n", *args, str(mock_video_path)]
    settings.update({"cache_dir": Path(tmp_path), "langs_to_keep": ["en"]})

    # And: Mocked video metadata and output path
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )
    mocker.patch("vid_cleaner.cli.clip_video.copy_file", return_value="clipped_video.mkv")

    # When: Running clip command in dry-run mode
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # THEN verify ffmpeg was not called
    mock_ffmpeg.assert_not_called()
    assert exc_info.value.code == 0
    assert "dry run" in output
    assert "✅ Success: clipped_video.mkv" not in output



"""Constants for vid-cleaner."""

import os
from dataclasses import dataclass
from enum import Enum, StrEnum
from pathlib import Path

PACKAGE_NAME = __package__.replace("_", "-").replace(".", "-").replace(" ", "-")
CONFIG_DIR = Path(os.getenv("XDG_CONFIG_HOME", "~/.config")).expanduser().absolute() / PACKAGE_NAME
DATA_DIR = Path(os.getenv("XDG_DATA_HOME", "~/.local/share")).expanduser().absolute() / PACKAGE_NAME
STATE_DIR = (
    Path(os.getenv("XDG_STATE_HOME", "~/.local/state")).expanduser().absolute() / PACKAGE_NAME
)
CACHE_DIR = Path(os.getenv("XDG_CACHE_HOME", "~/.cache")).expanduser().absolute() / PACKAGE_NAME
PROJECT_ROOT_PATH = Path(__file__).parents[2].absolute()
PACKAGE_ROOT_PATH = Path(__file__).parents[0].absolute()
USER_CONFIG_PATH = CONFIG_DIR / "config.toml"
DEFAULT_CONFIG_PATH = PACKAGE_ROOT_PATH / "default_config.toml"
DEV_DIR = PROJECT_ROOT_PATH / ".development"
DEV_CONFIG_PATH = DEV_DIR / "dev-config.toml"


class VideoContainerTypes(StrEnum):
    """Video container types for vid-cleaner."""

    MKV = ".mkv"
    MP4 = ".mp4"
    AVI = ".avi"
    WEBM = ".webm"
    MOV = ".mov"
    WMV = ".wmv"
    M4V = ".m4v"


class CodecTypes(StrEnum):
    """Codec types for vid-cleaner."""

    AUDIO = "audio"
    VIDEO = "video"
    SUBTITLE = "subtitle"
    ATTACHMENT = "attachment"
    DATA = "data"


class AudioLayout(Enum):
    """Audio layouts for vid-cleaner. Values are the number of streams."""

    MONO = 1
    STEREO = 2
    SURROUND5 = 6
    SURROUND7 = 8


class VideoTrait(StrEnum):
    """Video traits for vid-cleaner."""

    H265 = "h265"
    VP9 = "vp9"
    H264 = "h264"
    STEREO = "stereo"
    MONO = "mono"
    SURROUND5 = "surround5"
    SURROUND7 = "surround7"
    COMMENTARY = "commentary"
    NOSTEREO = "no_stereo"
    SURROUND_ONLY = "surround_only"
    FHD = "1080p"
    UHDTV = "4k"
    HDTV = "720p"
    SDTV = "480p"
    UNKNOWN_RESOLUTION = "unknown"
    REORDER = "reorder"

    @classmethod
    def help_options(cls) -> str:
        """Generate a formatted string of all available video trait values for help documentation.

        Create a comma-separated list of video trait values wrapped in backticks for use in command-line help text or documentation. This method provides a convenient way to display all possible trait options to users.

        Returns:
            str: A comma-separated string of video trait values formatted with backticks (e.g., "`h265`, `h264`, `stereo`").
        """
        values = [f"`{t.value}`" for t in cls]
        return ", ".join(values)


@dataclass
class Resolution:
    """Resolution for vid-cleaner."""

    width: int
    height: int


SYMBOL_CHECK = "✔"
COMMENTARY_STREAM_TITLE_REGEX = r"commentary|sdh|description"
EXCLUDED_VIDEO_CODECS = {"mjpeg", "mjpg", "png"}
FFMPEG_APPEND: list[str] = ["-max_muxing_queue_size", "9999"]
FFMPEG_PREPEND: list[str] = ["-y", "-hide_banner"]
H265_CODECS = {"hevc", "vp9"}
VERSION = "0.6.2"
FHD_RESOLUTION = Resolution(width=1920, height=1080)
UHDTV_RESOLUTION = Resolution(width=3840, height=2160)
HDTV_RESOLUTION = Resolution(width=1280, height=720)
SDTV_RESOLUTION = Resolution(width=720, height=480)

# how many bytes to read at once?
# shutil.copy uses 1024 * 1024 if _WINDOWS else 64 * 1024
# however, in my testing on MacOS with SSD, I've found a much larger buffer is faster
IO_BUFFER_SIZE = 4096 * 1024


class PrintLevel(Enum):
    """Define verbosity levels for console output.

    Use these levels to control the amount of information displayed to users. Higher levels include all information from lower levels plus additional details.
    """

    INFO = 0
    DEBUG = 1
    TRACE = 2



"""Tables for the VidCleaner application."""

from box import Box
from rich.table import Table


def stream_table(ffprobe_box: Box) -> Table:
    """Create a formatted table displaying video stream information.

    Display details about video, audio and subtitle streams in a terminal-friendly format. The table includes stream index, codec type, language, audio channels, dimensions and titles.

    Args:
        ffprobe_box (Box): Box object containing ffprobe output with stream information

    Returns:
        Table: Rich table containing formatted stream information
    """
    table = Table(title=ffprobe_box.name)
    table.add_column("#")
    table.add_column("Type")
    table.add_column("Codec Name")
    table.add_column("Language")
    table.add_column("Channels")
    table.add_column("Channel Layout")
    table.add_column("Width")
    table.add_column("Height")
    table.add_column("Title")

    for stream in ffprobe_box.streams:
        table.add_row(
            str(stream.index),
            stream.codec_type.value,
            stream.codec_name,
            stream.language,
            str(stream.channels.value) if stream.channels else "",
            stream.channel_layout or "",
            str(stream.width) if stream.width else "",
            str(stream.height) if stream.height else "",
            stream.title or "",
        )

    return table



"""Vidcleaner cli."""

from __future__ import annotations

from pathlib import Path
from typing import Annotated

import cappa
from nclutils import pp, print_debug
from rich.traceback import install

from vid_cleaner import settings
from vid_cleaner.config import SettingsManager
from vid_cleaner.constants import USER_CONFIG_PATH, PrintLevel, VideoTrait
from vid_cleaner.utils import create_default_config, parse_trait_filters


def config_subcommand(vidcleaner: VidCleaner) -> None:
    """Configure settings based on the provided command and arguments.

    Update the global settings object with values from the command line arguments and configuration file. Handle project-specific settings if a project is specified.

    Args:
        vidcleaner (VidCleaner): The main CLI application object containing command and configuration options.
    """
    pp.configure(
        debug=vidcleaner.verbosity in {PrintLevel.DEBUG, PrintLevel.TRACE},
        trace=vidcleaner.verbosity == PrintLevel.TRACE,
    )

    langs_to_keep = getattr(vidcleaner.command, "langs_to_keep", None)
    if langs_to_keep and isinstance(langs_to_keep, str):
        langs_to_keep = langs_to_keep.split(",")

    if getattr(vidcleaner.command, "filters", None):
        filters = parse_trait_filters(getattr(vidcleaner.command, "filters", None))
    else:
        filters = set()

    # Apply command-specific settings
    cli_settings = {
        "downmix_stereo": getattr(vidcleaner.command, "downmix_stereo", False),
        "drop_local_subs": getattr(vidcleaner.command, "drop_local_subs", False),
        "drop_original_audio": getattr(vidcleaner.command, "drop_original_audio", False),
        "dryrun": getattr(vidcleaner, "dryrun", False),
        "filters": filters,
        "force": getattr(vidcleaner.command, "force", False),
        "h265": getattr(vidcleaner.command, "h265", False),
        "keep_all_subtitles": getattr(vidcleaner.command, "keep_all_subtitles", False),
        "keep_commentary": getattr(vidcleaner.command, "keep_commentary", False),
        "keep_local_subtitles": getattr(vidcleaner.command, "keep_local_subtitles", False),
        "langs_to_keep": langs_to_keep,
        "out_path": getattr(vidcleaner.command, "out", None),
        "overwrite": getattr(vidcleaner.command, "overwrite", False),
        "save_each_step": getattr(vidcleaner.command, "save_each_step", False),
        "subcommand": vidcleaner.command.__class__.__name__.lower(),
        "video_1080": getattr(vidcleaner.command, "video_1080", False),
        "vp9": getattr(vidcleaner.command, "vp9", False),
    }

    SettingsManager.apply_cli_settings(cli_settings)

    if pp.is_trace:
        print_debug(
            custom=[
                {"Settings": settings.to_dict()},
                {"vidcleaner": vidcleaner.__dict__},
            ],
            envar_prefix="VIDCLEANER",
            packages=["cappa", "dynaconf", "rich", "nclutils"],
        )


@cappa.command(
    name="vidcleaner",
    description=f"""Transcode video files to different formats or configurations using ffmpeg. This script provides a simple CLI for common video transcoding tasks.

- **Inspect** video files to display detailed stream information
- **Clip** a section from a video file
- **Drop audio streams** containing undesired languages or commentary
- **Drop subtitles** containing undesired languages
- **Keep subtitles** if original audio is not in desired language
- **Downmix audio** to stereo
- **Convert** video files to H265 or VP9

The defaults can be overridden by using the various command line options or by editing the configuration file located at `{USER_CONFIG_PATH}`.

**Usage Examples:**
```shell
# Inspect video file:
vidcleaner inspect <video_file>

# Clip a one minute clip from a video file:
vidcleaner clip --start=00:00:00 --duration=00:01:00 <video_file>

#Transcode a video to H265 format and keep English audio
vidcleaner clean --h265 --langs=eng <video_file>

# Downmix audio to stereo and keep all subtitles
vidcleaner clean --downmix --keep-subs <video_file>
```
    """,
)
class VidCleaner:
    """Transcode video files to different formats or configurations using ffmpeg. This script provides a simple CLI for common video transcoding tasks."""

    command: cappa.Subcommands[
        CacheCommand | CleanCommand | InspectCommand | ClipCommand | SearchCommand
    ]

    verbosity: Annotated[
        PrintLevel,
        cappa.Arg(
            short=True,
            count=True,
            help="Verbosity level (`-v` or `-vv`)",
            choices=[],
            show_default=False,
            propagate=True,
        ),
    ] = PrintLevel.INFO

    dryrun: Annotated[
        bool,
        cappa.Arg(
            long=True,
            short="-n",
            help="Preview changes without modifying files",
            show_default=False,
            propagate=True,
        ),
    ] = False


@cappa.command(
    name="clean",
    invoke="vid_cleaner.cli.clean_video.main",
    help="Clean a video file",
    description=f"""\
Transcode video files to different formats or configurations.

Vidcleaner is versatile and allows for a range of transcoding options for video files with various options. You can select various audio and video settings, manage subtitles, and choose the output file format.

The defaults for this command will:

* Use English as the default language
* Drop commentary audio tracks
* Keep default language audio
* Keep original audio if it is not the default language
* Drop all subtitles unless the original audio is not your local language

Defaults for vid-cleaner are set in the configuration file located at `{USER_CONFIG_PATH}`. When vid-cleaner is run, it will create this file if it does not exist. All options can be overridden on the command line.

**NOTE:** If you've updated your user config file, the flags for the cli will work in reverse order. For example, if you've set `downmix_stereo = true` in your user config file, the flag `--downmix` will actually disable downmixing.

**Important:** Vid-cleaner makes decisions about which audio and subtitle tracks to keep based on the original language of the video. This is determined by querying the TMDb, Radarr, andSonarr APIs. To use this functionality, you must add the appropriate API keys to the configuration file.

**Usage Examples:**
```shell
# Transcode a video to H265 format and keep English audio:
vidcleaner clean --h265 --langs=eng <video_file>

# Downmix audio to stereo and keep all subtitles:
vidcleaner clean --downmix --keep-subs <video_file>
```
    """,
)
class CleanCommand:
    """Clean a video file."""

    files: Annotated[
        list[Path],
        cappa.Arg(help="Video file path(s)"),
    ]
    out: Annotated[
        Path | None,
        cappa.Arg(
            help="Output path (Default: `./<input_file>_1.xxx`)",
            long=True,
            short=True,
            show_default=False,
        ),
    ] = None
    overwrite: Annotated[
        bool,
        cappa.Arg(
            help="Do not create a backup of the original file if it would be overwritten",
            long=True,
            show_default=True,
        ),
    ] = False
    downmix_stereo: Annotated[
        bool,
        cappa.Arg(
            help="Create a stereo track if none exist",
            long="--downmix",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.downmix_stereo
    drop_original_audio: Annotated[
        bool,
        cappa.Arg(
            help="Drop original language audio if not specified in langs_to_keep",
            long="--drop-original",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.drop_original_audio
    keep_all_subtitles: Annotated[
        bool,
        cappa.Arg(
            help="Keep all subtitles",
            long="--keep-subs",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.keep_all_subtitles
    save_each_step: Annotated[
        bool,
        cappa.Arg(
            help="By default, the new video is saved after all steps are completed. Use this option to save the video after each step.",
            long="--save-each",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.save_each_step
    keep_commentary: Annotated[
        bool,
        cappa.Arg(
            help="Keep commentary audio",
            long="--keep-commentary",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.keep_commentary
    keep_local_subtitles: Annotated[
        bool,
        cappa.Arg(
            help="Keep subtitles matching the local language(s)",
            long="--keep-local-subs",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.keep_local_subtitles
    drop_local_subs: Annotated[
        bool,
        cappa.Arg(
            help="Force dropping local subtitles even if audio is not default language",
            long="--drop-local-subs",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.drop_local_subs
    langs_to_keep: Annotated[
        str,
        cappa.Arg(
            help="Languages to keep. Comma separated language ISO 639-1 codes (e.g. `en,es`)",
            long="--langs",
            show_default=True,
            group="Configuration",
        ),
    ] = settings.langs_to_keep
    h265: Annotated[
        bool,
        cappa.Arg(
            help="Convert to H265", long="--h265", show_default=True, group="Video Conversion"
        ),
    ] = False
    vp9: Annotated[
        bool,
        cappa.Arg(help="Convert to VP9", long="--vp9", show_default=True, group="Video Conversion"),
    ] = False
    video_1080: Annotated[
        bool,
        cappa.Arg(
            help="Convert to 1080p", long="--1080p", show_default=True, group="Video Conversion"
        ),
    ] = False
    force: Annotated[
        bool,
        cappa.Arg(
            help="Force processing of file even if it is already in the desired format",
            long="--force",
            show_default=True,
        ),
    ] = False


@cappa.command(
    name="inspect",
    invoke="vid_cleaner.cli.inspect_video.main",
    help="Inspect a video file",
    description="""\
Inspect video files to display detailed stream information.

Use this command to view detailed information about the video and audio streams of a video file. The information includes stream type, codec, language, and audio channel details. This command is useful for understanding the composition of a video file before performing operations like clipping or transcoding.
""",
)
class InspectCommand:
    """Inspect a video file."""

    files: Annotated[
        list[Path],
        cappa.Arg(help="Video file path(s)"),
    ]
    json_output: Annotated[
        bool,
        cappa.Arg(
            help="Output in JSON format",
            long="--json",
            short=True,
            show_default=True,
        ),
    ] = False


@cappa.command(
    name="clip",
    invoke="vid_cleaner.cli.clip_video.main",
    help="Clip a video file",
    description="""\
Clip a section from a video file.

This command allows you to extract a specific portion of a video file based on start time and duration.

* The start time and duration should be specified in `HH:MM:SS` format.
* You can also specify an output file to save the clipped video. If the output file is not specified, the clip will be saved with a default naming convention.

Use the `--overwrite` option to avoid creating a backup of the original file if it would be overwritten.
""",
)
class ClipCommand:
    """Clip a video file."""

    files: Annotated[
        list[Path],
        cappa.Arg(help="Video file path(s)"),
    ]

    start: Annotated[
        str,
        cappa.Arg(
            help="Start time in `HH:MM:SS` format (Default: `00:00:00`)",
            long=True,
            short=True,
            show_default=False,
        ),
    ] = "00:00:00"
    duration: Annotated[
        str,
        cappa.Arg(
            help="Duration in `HH:MM:SS` format (Default: `00:01:00`)",
            long=True,
            short=True,
            show_default=False,
        ),
    ] = "00:01:00"
    out: Annotated[
        Path | None,
        cappa.Arg(
            help="Output file path (Default: `<input_file>_1`)",
            long=True,
            short=True,
            show_default=False,
        ),
    ] = None
    overwrite: Annotated[
        bool,
        cappa.Arg(
            help="Do not create a backup of the original file if it would be overwritten",
            long=True,
            show_default=True,
        ),
    ] = False


@cappa.command(
    name="cache",
    help="View and clear the vidcleaner cache",
    invoke="vid_cleaner.cli.cache.main",
)
class CacheCommand:
    """Manage the vidcleaner cache."""

    clear: Annotated[
        bool,
        cappa.Arg(help="Clear the vidcleaner cache", long=True, short=True, show_default=True),
    ] = False


@cappa.command(
    name="search",
    help="Search for video files under a directory",
    description="""\
This command allows you to search for video files under a directory and display detailed information about the video and audio streams of a video file. The information includes stream type, codec, language, and audio channel details. This command is useful for understanding the composition of a video file before performing operations like clipping or transcoding.

By using filters, you can search for video files that match specific criteria. For example, you can search for video files that are in the H264 codec and have a resolution of 1080p.

**Usage Examples:**
```shell
# Search for video files that are in the H264 codec and have a resolution of 1080p up to 2 levels deep:
vidcleaner search --filters=h264,1080p --depth=2
```
""",
    invoke="vid_cleaner.cli.search.main",
)
class SearchCommand:
    """Search for video files under a directory."""

    directory: Annotated[
        Path,
        cappa.Arg(help="Directory to search for video files", show_default=False),
    ] = Path.cwd()
    depth: Annotated[
        int,
        cappa.Arg(
            help="Depth to search for video files", long=True, short=False, show_default=False
        ),
    ] = 0
    filters: Annotated[
        str,
        cappa.Arg(
            help=f"Comma separated list of facets to search for. Valid options: {VideoTrait.help_options()}",
            long=True,
            short=False,
            show_default=False,
        ),
    ] = None


def main() -> None:  # pragma: no cover
    """Main function."""  # noqa: DOC501
    install(show_locals=True)

    try:
        cappa.invoke(obj=VidCleaner, deps=[create_default_config, config_subcommand])
    except KeyboardInterrupt as e:
        pp.info("Exiting...")
        raise cappa.Exit(code=1) from e


if __name__ == "__main__":  # pragma: no cover
    main()



"""Clip subcommand."""

import re

import cappa
from nclutils import copy_file, pp

from vid_cleaner import settings
from vid_cleaner.utils import coerce_video_files
from vid_cleaner.vidcleaner import ClipCommand


def main(clip_cmd: ClipCommand) -> None:
    """Extract video clips based on start time and duration.

    Create video clips by copying a section of the source video without re-encoding. Useful for extracting highlights or samples from longer videos.

    Args:
        cmd (VidCleaner): Global command options and configuration
        clip_cmd (ClipCommand): Clip-specific command options

    Raises:
        cappa.Exit: If start or duration times are not in HH:MM:SS format
    """
    time_pattern = re.compile(r"^\d{2}:\d{2}:\d{2}$")

    if not time_pattern.match(clip_cmd.start):
        pp.error("`--start` must be in format HH:MM:SS")  # type: ignore [unreachable]
        raise cappa.Exit(code=1)

    if not time_pattern.match(clip_cmd.duration):
        pp.error("`--duration` must be in format HH:MM:SS")  # type: ignore [unreachable]
        raise cappa.Exit(code=1)

    for video in coerce_video_files(clip_cmd.files):
        settings.out_path = settings.out_path or video.path
        pp.info(f"⇨ {video.path.name}")

        video.clip(clip_cmd.start, clip_cmd.duration)

        if not settings.dryrun:
            out_file = copy_file(
                src=video.temp_file.latest_temp_path(),
                dst=settings.out_path,
                keep_backup=not settings.overwrite,
                with_progress=True,
                transient=True,
            )
            video.temp_file.clean_up()
            pp.success(f"{out_file}")

    raise cappa.Exit(code=0)



"""API utilities."""

import requests
from nclutils import console, pp

from vid_cleaner import settings


def query_tmdb(search: str) -> dict:  # pragma: no cover
    """Query The Movie Database API for a movie title.

    Args:
        search (str): IMDB id (tt____) to search for

    Returns:
        dict: The Movie Database API response
    """
    tmdb_api_key = settings.TMDB_API_KEY

    if not tmdb_api_key:
        return {}

    url = f"https://api.themoviedb.org/3/find/{search}"

    params = {
        "api_key": tmdb_api_key,
        "language": "en-US",
        "external_source": "imdb_id",
    }

    if pp.is_trace:
        args = "&".join([f"{k}={v}" for k, v in params.items()])
        pp.trace(f"TMDB: Querying {url}?{args}")

    try:
        response = requests.get(url, params=params, timeout=15)
    except Exception as e:  # noqa: BLE001
        pp.error(str(e))
        return {}

    if response.status_code != 200:  # noqa: PLR2004
        pp.error(
            f"Error querying The Movie Database API: {response.status_code} {response.reason}",
        )
        return {}

    pp.trace("TMDB: Response received")
    if pp.is_trace:
        console.log(response.json())
    return response.json()


def query_radarr(search: str) -> dict:  # pragma: no cover
    """Query Radarr API for a movie title.

    Args:
        search (str): Movie title to search for
        api_key (str): Radarr API key

    Returns:
        dict: Radarr API response
    """
    radarr_url = settings.RADARR_URL
    radarr_api_key = settings.RADARR_API_KEY

    if not radarr_api_key or not radarr_url:
        return {}

    url = f"{radarr_url}/api/v3/parse"
    params = {
        "apikey": radarr_api_key,
        "title": search,
    }

    try:
        response = requests.get(url, params=params, timeout=15)
    except Exception as e:  # noqa: BLE001
        pp.error(str(e))
        return {}

    if response.status_code != 200:  # noqa: PLR2004
        pp.error(f"Error querying Radarr: {response.status_code} {response.reason}")
        return {}

    return response.json()


def query_sonarr(search: str) -> dict:  # pragma: no cover
    """Query Sonarr API for a movie title.

    Args:
        search (str): Movie title to search for
        api_key (str): Radarr API key

    Returns:
        dict: Sonarr API response
    """
    sonarr_url = settings.SONARR_URL
    sonarr_api_key = settings.SONARR_API_KEY

    if not sonarr_api_key or not sonarr_url:
        return {}

    url = f"{sonarr_url}/api/v3/parse"
    params = {
        "apikey": sonarr_api_key,
        "title": search,
    }

    try:
        response = requests.get(url, params=params, timeout=15)
    except Exception as e:  # noqa: BLE001
        pp.error(str(e))
        return {}

    if response.status_code != 200:  # noqa: PLR2004
        pp.error(f"Error querying Sonarr: {response.status_code} {response.reason}")
        return {}

    pp.trace("SONARR: Response received")
    return response.json()



"""Test TempFile controller functionality."""

from pathlib import Path

from vid_cleaner import settings

from vid_cleaner.controllers import TempFile  # isort: skip


def test_temp_file_latest_temp_path(mock_video_path: Path, tmp_path: Path) -> None:
    """Verify original path is returned when no temp files exist."""
    # Given: A TempFile instance with cache directory set
    settings.update({"cache_dir": Path(tmp_path)})
    temp_file = TempFile(mock_video_path)

    # When: Getting latest temp path with no temp files
    result = temp_file.latest_temp_path()

    # Then: Original path is returned
    assert result == mock_video_path


def test_temp_file_new_tmp_path(mock_video_path: Path, tmp_path: Path) -> None:
    """Verify new temp file paths are generated with optional suffix and step name."""
    # Given: A TempFile instance with cache directory set
    settings.update({"cache_dir": Path(tmp_path)})
    temp_file = TempFile(mock_video_path)

    # When: Generating new temp paths with different options
    default_path = temp_file.new_tmp_path()
    custom_path = temp_file.new_tmp_path(suffix=".mkv", step_name="somename")

    # Then: Paths are generated with correct naming
    assert default_path.name == "1_no_step.mp4"
    assert custom_path.name == "1_somename.mkv"
    assert temp_file.created_tmp_files == []


def test_temp_file_created_temp_file(mock_video_path: Path, tmp_path: Path) -> None:
    """Verify newly created temp files are tracked correctly."""
    # Given: A TempFile instance with cache directory set
    settings.update({"cache_dir": Path(tmp_path)})
    temp_file = TempFile(mock_video_path)

    # When: Creating and tracking a new temp file
    output_path = temp_file.new_tmp_path(suffix=".mkv", step_name="somename")
    temp_file.created_temp_file(output_path)

    # Then: File is tracked in created_tmp_files
    assert temp_file.created_tmp_files == [output_path]


def test_temp_file_clean_old_tmp_files(mock_video_path: Path, tmp_path: Path) -> None:
    """Verify all but latest temp file are removed."""
    # Given: A TempFile instance with multiple temp files
    settings.update({"cache_dir": Path(tmp_path)})
    temp_file = TempFile(mock_video_path)

    # When: Creating multiple temp files
    output_path_1 = temp_file.new_tmp_path()
    output_path_1.touch()
    temp_file.created_temp_file(output_path_1)

    output_path_2 = temp_file.new_tmp_path()
    output_path_2.touch()
    temp_file.created_temp_file(output_path_2)

    output_path_3 = temp_file.new_tmp_path()
    output_path_3.touch()
    temp_file.created_temp_file(output_path_3)

    # When: Cleaning old temp files
    temp_file.clean_old_tmp_files()

    # Then: Only latest file remains
    assert temp_file.created_tmp_files == [output_path_1, output_path_2, output_path_3]
    assert not output_path_1.exists()
    assert not output_path_2.exists()
    assert output_path_3.exists()


def test_temp_file_clean_up(mock_video_path: Path, tmp_path: Path) -> None:
    """Verify all temp files and temp directory are removed."""
    # Given: A TempFile instance with multiple temp files
    settings.update({"cache_dir": Path(tmp_path)})
    temp_file = TempFile(mock_video_path)

    # When: Creating multiple temp files
    output_path_1 = temp_file.new_tmp_path()
    output_path_1.touch()
    assert output_path_1.name.startswith("1_")
    temp_file.created_temp_file(output_path_1)

    output_path_2 = temp_file.new_tmp_path()
    output_path_2.touch()
    assert output_path_2.name.startswith("2_")
    temp_file.created_temp_file(output_path_2)

    output_path_3 = temp_file.new_tmp_path()
    output_path_3.touch()
    assert output_path_3.name.startswith("3_")
    temp_file.created_temp_file(output_path_3)

    # When: Cleaning up all temp files
    temp_file.clean_up()

    # Then: All temp files and directory are removed
    assert not output_path_1.exists()
    assert not output_path_2.exists()
    assert not output_path_3.exists()
    assert not output_path_3.parent.exists()



"""Manage temporary files."""

import uuid
from pathlib import Path

from nclutils import pp

from vid_cleaner import settings


class TempFile:
    """Manage temporary files created during video processing.

    This class handles the creation, tracking, and cleanup of temporary files generated during video processing operations. It maintains a unique temporary directory for each instance and manages file naming and cleanup.

    Args:
        path (Path): The path to the original video file.

    Attributes:
        path (Path): The resolved path to the original video file
        suffix (str): The file extension of the original video file
        tmp_dir (Path): Path to the unique temporary directory for this instance
        created_tmp_files (list[Path]): List tracking all temporary files created
        tmp_file_number (int): Counter for naming temporary files
    """

    def __init__(self, path: Path) -> None:
        """Initialize the TempFile controller.

        Args:
            path (Path): Path to the original video file to process
        """
        self.path = path.expanduser().resolve()
        self.suffix = path.suffix
        self.tmp_dir = settings.CACHE_DIR / uuid.uuid4().hex
        self.created_tmp_files: list[Path] = []
        self.tmp_file_number = 1

    def latest_temp_path(self) -> Path:
        """Get the path to the most recently created temporary file.

        Returns the path to the most recent temporary file that was created. If no
        temporary files exist yet, returns the path to the original input file.

        Returns:
            Path: Path to the most recent temporary file or original file
        """
        return self.created_tmp_files[-1] if self.created_tmp_files else self.path

    def new_tmp_path(self, suffix: str = "", step_name: str = "") -> Path:
        """Generate a new tmp path for the creation of a new temporary file.

        Creates a new unique path for a temporary file in the temp directory. The path
        will include an incrementing number and optional step name in the filename.

        Args:
            suffix (str | None, optional): File extension to use. Defaults to original file's suffix.
            step_name (str | None, optional): Name of processing step to include in filename. Defaults to None.

        Returns:
            Path: Path object for the new temporary file
        """
        # Get the output file name
        suffix = suffix or self.suffix
        self.tmp_dir.mkdir(parents=True, exist_ok=True)

        if not step_name:
            step_name = "no_step"

        # Create a new tmp file name
        for file in self.tmp_dir.iterdir():
            if file.stem.startswith(f"{self.tmp_file_number}_"):
                self.tmp_file_number += 1
        pp.trace(f"Create new tmp file: {self.tmp_file_number}_{step_name}{suffix}")
        return self.tmp_dir / f"{self.tmp_file_number}_{step_name}{suffix}"

    def created_temp_file(self, path: Path) -> None:
        """Register a newly created temporary file.

        Adds the given path to the list of created temporary files and triggers cleanup
        of older temporary files to manage disk space.

        Args:
            path (Path): Path to the newly created temporary file
        """
        self.created_tmp_files.append(path)
        # Always clean old temporary files to minimize the size of the tmp directory
        self.clean_old_tmp_files()

    def clean_old_tmp_files(self) -> None:
        """Remove older temporary files.

        Deletes all temporary files except the most recent one to conserve disk space. Files are identified as "old" if their number is less than the current counter.
        """
        if self.tmp_dir.exists():
            for file in self.tmp_dir.iterdir():
                tmp_file_number = int(file.stem.split("_")[0])
                if tmp_file_number < self.tmp_file_number:
                    pp.trace(f"Remove tmp file: {file}")
                    file.unlink()

    def clean_up(self) -> None:
        """Remove all temporary files and the temporary directory.

        Performs complete cleanup by removing all temporary files created during
        processing and the temporary directory itself.
        """
        if self.tmp_dir.exists():
            pp.debug("Clean up temporary files")

            for file in self.tmp_dir.iterdir():
                pp.trace(f"Remove tmp file: {file}")
                file.unlink()

            pp.trace(f"Remove: {self.tmp_dir}")
            self.tmp_dir.rmdir()



"""VideoFile model."""

import atexit
import re
from pathlib import Path
from typing import assert_never

import cappa
from box import Box
from ffmpeg_progress_yield import FfmpegProgress
from iso639 import Lang
from nclutils import console, pp
from rich.markdown import Markdown
from rich.progress import Progress

from vid_cleaner import settings
from vid_cleaner.constants import (
    COMMENTARY_STREAM_TITLE_REGEX,
    EXCLUDED_VIDEO_CODECS,
    FFMPEG_APPEND,
    FFMPEG_PREPEND,
    FHD_RESOLUTION,
    H265_CODECS,
    HDTV_RESOLUTION,
    SDTV_RESOLUTION,
    SYMBOL_CHECK,
    UHDTV_RESOLUTION,
    AudioLayout,
    CodecTypes,
    VideoTrait,
)
from vid_cleaner.utils import get_probe_as_box, query_radarr, query_sonarr, query_tmdb, run_ffprobe

from vid_cleaner.controllers import TempFile  # isort: skip


def cleanup_on_exit(video_file: "VideoFile") -> None:  # pragma: no cover
    """Cleanup temporary files on exit.

    Args:
        video_file (VideoFile): The VideoFile object to perform cleanup on.
    """
    video_file.temp_file.clean_up()


class VideoFile:
    """VideoFile model."""

    def __init__(self, path: Path) -> None:
        """Initialize VideoFile."""
        self.path = path.expanduser().resolve()
        self.name = path.name
        self.stem = path.stem
        self.parent = path.parent
        self.suffix = path.suffix
        self.suffixes = self.path.suffixes
        self.temp_file = TempFile(self.path)

        self.container = self.suffix
        self.language: Lang = None
        self.ran_language_check = False
        self._probe_box: Box = Box({}, default_box=True, default_box_create_on_get=False)

        self._video_streams: list[Box] = []
        self._audio_streams: list[Box] = []
        self._subtitle_streams: list[Box] = []

        atexit.register(cleanup_on_exit, self)

    @property
    def probe_box(self) -> Box:
        """Get the probe box."""
        if self._probe_box.path_to_file != self.temp_file.latest_temp_path():
            self._probe_box = get_probe_as_box(self.temp_file.latest_temp_path())

        return self._probe_box

    @property
    def video_streams(self) -> list[Box]:
        """Get the video streams."""
        if not self._video_streams:
            self._video_streams = [
                s
                for s in self.probe_box.streams
                if s.codec_type == CodecTypes.VIDEO
                and s.codec_name.lower() not in EXCLUDED_VIDEO_CODECS
            ]
        return self._video_streams

    @property
    def audio_streams(self) -> list[Box]:
        """Get the audio streams."""
        if not self._audio_streams:
            self._audio_streams = [
                s for s in self.probe_box.streams if s.codec_type == CodecTypes.AUDIO
            ]
        return self._audio_streams

    @property
    def subtitle_streams(self) -> list[Box]:
        """Get the subtitle streams."""
        if not self._subtitle_streams:
            self._subtitle_streams = [
                s for s in self.probe_box.streams if s.codec_type == CodecTypes.SUBTITLE
            ]
        return self._subtitle_streams

    def get_traits(self) -> list[VideoTrait]:
        """Analyze video file streams to identify audio, video, and structural characteristics.

        Extract comprehensive traits from the video file by examining audio streams for channel layouts and commentary tracks, video streams for codec types and resolutions, and derive additional traits based on stream ordering requirements and audio configuration gaps.

        Returns:
            list[VideoTrait]: A list of VideoTrait enums representing all identified characteristics including audio layouts, video codecs, resolutions, and structural properties.
        """
        traits = []

        # Process audio streams
        traits.extend(self._get_audio_traits())

        # Process video streams
        traits.extend(self._get_video_traits())

        # Add derived traits
        if VideoTrait.STEREO not in traits:
            traits.append(VideoTrait.NOSTEREO)

        if VideoTrait.STEREO not in traits and VideoTrait.MONO not in traits:
            traits.append(VideoTrait.SURROUND_ONLY)

        if self._need_stream_reorder():
            traits.append(VideoTrait.REORDER)

        return traits

    def _get_audio_traits(self) -> list[VideoTrait]:
        """Extract audio-related traits from the video file's audio streams.

        Analyze each audio stream to identify characteristics such as channel layout
        (stereo, mono, surround sound) and special properties like commentary tracks.
        Commentary tracks are identified by matching stream titles against a regex pattern.

        Returns:
            list[VideoTrait]: A list of VideoTrait enums representing the audio characteristics
                found in the file's audio streams.
        """
        traits = []
        for stream in self.audio_streams:
            if stream.title and re.search(
                COMMENTARY_STREAM_TITLE_REGEX, stream.title, re.IGNORECASE
            ):
                traits.append(VideoTrait.COMMENTARY)
            elif stream.channels == AudioLayout.STEREO:
                traits.append(VideoTrait.STEREO)
            elif stream.channels == AudioLayout.MONO:
                traits.append(VideoTrait.MONO)
            elif stream.channels == AudioLayout.SURROUND5:
                traits.append(VideoTrait.SURROUND5)
            elif stream.channels == AudioLayout.SURROUND7:
                traits.append(VideoTrait.SURROUND7)
        return traits

    def _get_video_traits(self) -> list[VideoTrait]:
        """Extract video-related traits from the video file's video streams.

        Analyze each video stream to identify codec type (H.264, H.265) and resolution
        characteristics (HDTV, FHD, UHDTV, SDTV). Resolution is determined by comparing
        both height and width against standard resolution constants.

        Returns:
            list[VideoTrait]: A list of VideoTrait enums representing the video characteristics
                found in the file's video streams, including codec and resolution information.
        """
        traits = []
        for stream in self.video_streams:
            if stream.codec_name.lower() in H265_CODECS:
                traits.append(VideoTrait.H265)
            elif stream.codec_name.lower() == "h264":
                traits.append(VideoTrait.H264)

            if stream.height == HDTV_RESOLUTION.height or stream.width == HDTV_RESOLUTION.width:
                traits.append(VideoTrait.HDTV)
            elif stream.height == FHD_RESOLUTION.height or stream.width == FHD_RESOLUTION.width:
                traits.append(VideoTrait.FHD)
            elif stream.height == UHDTV_RESOLUTION.height or stream.width == UHDTV_RESOLUTION.width:
                traits.append(VideoTrait.UHDTV)
            elif stream.height == SDTV_RESOLUTION.height or stream.width == SDTV_RESOLUTION.width:
                traits.append(VideoTrait.SDTV)
            else:
                traits.append(VideoTrait.UNKNOWN_RESOLUTION)
        return traits

    @staticmethod
    def _downmix_to_stereo(streams: list[Box]) -> list[str]:
        """Generate a partial ffmpeg command to downmix audio streams to stereo if needed.

        Analyze the provided audio streams and construct a command to downmix 5.1 or 7.1 audio streams to stereo. Handle cases where stereo is already present or needs to be created from surround sound streams.

        Args:
            streams (list[Box]): List of audio stream dictionaries.

        Returns:
            list[str]: A list of strings forming part of an ffmpeg command for audio downmixing.
        """
        downmix_command: list[str] = []
        new_index = 0
        has_stereo = False
        surround5 = []  # Track 5.1 streams for potential downmixing
        surround7 = []  # Track 7.1 streams for potential downmixing

        for stream in streams:
            match stream.channels:
                case AudioLayout.STEREO:
                    has_stereo = True
                case AudioLayout.SURROUND5:
                    surround5.append(stream)
                case AudioLayout.SURROUND7:
                    surround7.append(stream)
                case AudioLayout.MONO:
                    pass
                case _:
                    assert_never(stream.channels)

        if not has_stereo and surround5:
            for surround5_stream in surround5:
                # Custom pan filter to preserve center channel dialogue and add LFE for bass impact
                # Coefficients tuned to maintain dialogue clarity while preserving surround ambiance
                downmix_command.extend(
                    [
                        "-map",
                        f"0:{surround5_stream.index}",
                        f"-c:a:{new_index}",
                        "aac",
                        f"-ac:a:{new_index}",
                        "2",
                        f"-b:a:{new_index}",
                        "256k",
                        f"-filter:a:{new_index}",
                        "pan=stereo|FL=FC+0.30*FL+0.30*FLC+0.30*BL+0.30*SL+0.60*LFE|FR=FC+0.30*FR+0.30*FRC+0.30*BR+0.30*SR+0.60*LFE,loudnorm",
                        f"-ar:a:{new_index}",
                        "48000",
                        f"-metadata:s:a:{new_index}",
                        "title=2.0",
                    ],
                )
                new_index += 1
                has_stereo = True

        if not has_stereo and surround7:
            pp.debug(
                "PROCESS AUDIO: Audio track is 5 channel, no 2 channel exists. Creating 2 channel from 5 channel",
            )
            # For 7.1, use default ffmpeg downmixing since custom pan filter would be too complex
            # and the default algorithm provides good results for 7.1 sources
            for surround7_stream in surround7:
                downmix_command.extend(
                    [
                        "-map",
                        f"0:{surround7_stream.index}",
                        f"-c:a:{new_index}",
                        "aac",
                        f"-ac:a:{new_index}",
                        "2",
                        f"-b:a:{new_index}",
                        "256k",
                        f"-metadata:s:a:{new_index}",
                        "title=2.0",
                    ],
                )
                new_index += 1

        pp.trace(f"PROCESS AUDIO: Downmix command: {downmix_command}")
        return downmix_command

    def _find_original_language(self) -> Lang:  # pragma: no cover
        """Find the original language of the video content.

        Query external APIs (IMDb, TMDB, Radarr, Sonarr) to determine the original language. Cache results to avoid repeated API calls.

        Returns:
            Lang: The determined original language code
        """
        if self.ran_language_check:
            return self.language

        original_language = None

        # Extract IMDb ID from filename if present (e.g. "Movie.Title.tt1234567.mkv")
        match = re.search(r"(tt\d+)", self.stem)
        imdb_id = match.group(0) if match else self._query_arr_apps_for_imdb_id()

        response = query_tmdb(imdb_id) if imdb_id else None

        if response and (tmdb_response := response.get("movie_results", [{}])[0]):
            original_language = tmdb_response.get("original_language")
            pp.trace(f"TMDB: Original language: {original_language}")

        if not original_language:
            pp.debug(f"Could not find original language for: {self.name}")
            return None

        # TMDB uses 'cn' for Chinese but iso639 expects 'zh'
        if original_language == "cn":
            original_language = "zh"

        try:
            language = Lang(original_language)
        except Exception:  # noqa: BLE001
            pp.debug(f"iso639: Could not find language for: {self.name}")
            return None

        self.language = language
        self.ran_language_check = True
        return language

    def _need_stream_reorder(self) -> bool:
        """Check if the video file needs stream reordering.

        Returns:
            bool: True if the video file needs stream reordering, False otherwise.
        """
        return any(
            stream.index != i
            for i, stream in enumerate(
                self.video_streams + self.audio_streams + self.subtitle_streams
            )
        )

    def _process_audio(self) -> tuple[list[str], list[str]]:
        """Construct commands for processing audio streams.

        Analyze and process audio streams based on language, commentary, and downmixing criteria. Generate ffmpeg commands for keeping or altering audio streams as required.

        Returns:
            tuple[list[str], list[str]]: A tuple containing two lists of strings forming part of an ffmpeg command for audio processing.
        """
        command: list[str] = []

        langs = [Lang(lang) for lang in settings.langs_to_keep]

        # Add original language to list of languages to keep if not explicitly dropping it
        if not settings.drop_original_audio:
            original_language = self._find_original_language()
            if original_language and original_language not in langs:
                langs.append(original_language)

        streams_to_keep = []
        for stream in self.audio_streams:
            # Unknown language streams are kept to avoid removing potentially important audio
            if not stream.language:
                command.extend(["-map", f"0:{stream.index}"])
                streams_to_keep.append(stream)
                continue

            # Commentary tracks are often unwanted and take up space
            if (
                not settings.keep_commentary
                and stream.title
                and re.search(COMMENTARY_STREAM_TITLE_REGEX, stream.title, re.IGNORECASE)
            ):
                pp.trace(rf"PROCESS AUDIO: Remove stream #{stream.index} \[commentary]")
                continue

            if stream.language == "und" or Lang(stream.language) in langs:
                command.extend(["-map", f"0:{stream.index}"])
                streams_to_keep.append(stream)
                continue

            pp.trace(f"PROCESS AUDIO: Remove stream #{stream.index}")

        # If all streams would be removed, keep them all to prevent silent video
        if not command:
            for stream in self.audio_streams:
                command.extend(["-map", f"0:{stream.index}"])
                streams_to_keep.append(stream)

        # Create stereo downmix commands if requested
        downmix_command = (
            self._downmix_to_stereo(streams_to_keep) if settings.downmix_stereo else []
        )

        pp.trace(f"PROCESS AUDIO: {command}")
        return command, downmix_command

    def _process_subtitles(self) -> list[str]:
        """Construct a command list for processing subtitle streams.

        Analyze and filter subtitle streams based on language preferences, commentary options, and other criteria. Build an ffmpeg command list accordingly.

        Returns:
            list[str]: A list of strings forming part of an ffmpeg command for subtitle processing.
        """
        command: list[str] = []

        langs = [Lang(lang) for lang in settings.langs_to_keep]

        # Only look up original language if we're not explicitly dropping local subs
        # This avoids unnecessary API calls
        if not settings.drop_local_subs:
            original_language = self._find_original_language()

        # Early return if no subtitle streams should be kept based on settings
        if (
            not settings.keep_all_subtitles
            and not settings.keep_local_subtitles
            and settings.drop_local_subs
        ):
            return command

        for stream in self.subtitle_streams:
            # Remove commentary/SDH/description tracks unless explicitly kept
            # These are typically supplementary and take up extra space
            if (
                not settings.keep_commentary
                and stream.title is not None
                and re.search(COMMENTARY_STREAM_TITLE_REGEX, stream.title, re.IGNORECASE)
            ):
                pp.trace(rf"PROCESS SUBTITLES: Remove stream #{stream.index} \[commentary]")
                continue

            if settings.keep_all_subtitles:
                command.extend(["-map", f"0:{stream.index}"])
                continue

            if stream.language:
                # Keep undefined language streams and streams matching user preferences
                # This ensures we don't accidentally remove important subtitles
                if settings.keep_local_subtitles and (
                    stream.language.lower() == "und" or Lang(stream.language) in langs
                ):
                    pp.trace(f"PROCESS SUBTITLES: Keep stream #{stream.index} (local language)")
                    command.extend(["-map", f"0:{stream.index}"])
                    continue

                # Keep subtitles in user's languages when original audio differs
                # This ensures subtitles are available when needed for translation
                if (
                    not settings.drop_local_subs
                    and langs
                    and original_language not in langs
                    and (stream.language.lower == "und" or Lang(stream.language) in langs)
                ):
                    pp.trace(f"PROCESS SUBTITLES: Keep stream #{stream.index} (original language)")
                    command.extend(["-map", f"0:{stream.index}"])
                    continue

            pp.trace(f"PROCESS SUBTITLES: Remove stream #{stream.index}")

        pp.trace(f"PROCESS SUBTITLES: {command}")
        return command

    def _process_video(self) -> list[str]:
        """Create a command list for processing video streams.

        Iterate through the provided video streams and construct a list of ffmpeg commands to process them, excluding any streams with codecs in the exclusion list.

        Returns:
            list[str]: A list of strings forming part of an ffmpeg command for video processing.
        """
        command: list[str] = []
        for stream in self.video_streams:
            if stream.codec_name.lower() in EXCLUDED_VIDEO_CODECS:
                continue

            command.extend(["-map", f"0:{stream.index}"])

        pp.trace(f"PROCESS VIDEO: {command}")
        return command

    def _query_arr_apps_for_imdb_id(self) -> str | None:
        """Query Radarr and Sonarr APIs to find the IMDb ID of the video.

        This method attempts to retrieve the IMDb ID based on the video file's name by utilizing external APIs for Radarr and Sonarr as sources. It first queries Radarr API and checks if the response contains the movie information with the IMDb ID. If found, it returns the IMDb ID.

        If not found, it then queries Sonarr API and checks if the response contains the series information with the IMDb ID. If found, it returns the IMDb ID. If no IMDb ID is found from either API, it returns None.

        Returns:
            str | None: The IMDb ID if found, otherwise None.
        """
        response = query_radarr(self.name)
        if response and "movie" in response and "imdbId" in response["parsedMovieInfo"]:
            return response["movie"]["imdbId"]

        response = query_sonarr(self.name)
        if response and "series" in response and "imdbId" in response["series"]:
            return response["series"]["imdbId"]

        return None

    def _run_ffmpeg(
        self,
        command: list[str],
        title: str,
        suffix: str | None = None,
        step: str | None = None,
    ) -> Path:
        """Execute an ffmpeg command.

        Run the provided ffmpeg command, showing progress and logging information. Determine input and output paths, and manage temporary files related to the operation.

        Args:
            command (list[str]): The ffmpeg command to execute.
            title (str): Title for logging the process.
            suffix (str | None, optional): Suffix for the output file. Use when creating a new container mime type. Defaults to None.
            step (str | None, optional): Step name for file naming. Used when creating a new temporary file. Defaults to None.

        Returns:
            Path: Path to the output file generated by the ffmpeg command.

        Raises:
            cappa.Exit: If KeyboardInterrupt occurs during the ffmpeg command.
        """
        input_path = self.temp_file.latest_temp_path()
        output_path = self.temp_file.new_tmp_path(suffix=suffix, step_name=step)

        # Prepend global ffmpeg options before input file to ensure consistent behavior
        cmd: list[str] = ["ffmpeg", *FFMPEG_PREPEND, "-i", str(input_path)]
        cmd.extend(command)
        cmd.extend([*FFMPEG_APPEND, str(output_path)])

        pp.trace(f"RUN FFMPEG:\n{' '.join(cmd)}")

        if settings.dryrun:
            console.rule(f"{title} (dry run)")
            markdown_command = Markdown(f"```console\n{' '.join(cmd)}\n```")
            console.print(markdown_command)
            return output_path

        # Use FfmpegProgress to get real-time progress updates during encoding
        ff = FfmpegProgress(cmd)

        try:
            with Progress(transient=True) as progress:
                task = progress.add_task(f"{title}…", total=100)
                for complete in ff.run_command_with_progress():
                    progress.update(task, completed=complete)
        except KeyboardInterrupt as e:
            # Clean up temporary files if user interrupts to avoid orphaned files
            self.temp_file.clean_up()
            pp.warning(f"KeyboardInterrupt during {title.lower()}")
            pp.info("Exiting...")
            raise cappa.Exit(code=1) from e

        pp.info(f"{SYMBOL_CHECK} {title}")

        self.temp_file.created_temp_file(output_path)
        pp.trace(f"Created temp file: {output_path}")
        return output_path

    def clip(self, start: str, duration: str) -> Path:
        """Clip a segment from the video.

        Extract a specific portion of the video based on the given start time and duration. Utilize ffmpeg to perform the clipping operation.

        Args:
            start (str): Start time of the clip.
            duration (str): Duration of the clip.

        Returns:
            Path: Path to the clipped video file.
        """
        ffmpeg_command: list[str] = ["-ss", start, "-t", duration, "-map", "0", "-c", "copy"]

        return self._run_ffmpeg(ffmpeg_command, title="Clip video", step="clip")

    def convert_to_h265(self) -> Path:
        """Convert the video to H.265 codec format.

        Check if conversion is necessary and perform it if so. This involves calculating the bitrate, building the ffmpeg command, and running it. Return the path to the converted video or the original video if conversion isn't needed.

        Returns:
            Path: Path to the converted or original video file.
        """
        input_path = self.temp_file.latest_temp_path()

        video_stream = next(
            stream
            for stream in self.probe_box.streams
            if stream.codec_type == CodecTypes.VIDEO
            and stream.codec_name.lower() not in EXCLUDED_VIDEO_CODECS
        )

        if not video_stream:
            pp.error("No video stream found")
            return input_path

        if not settings.force and video_stream.codec_name.lower() in H265_CODECS:
            pp.warning(
                "H265 ENCODE: Video already H.265 or VP9. Run with `--force` to re-encode. Skipping",
            )
            return input_path

        # Calculate target bitrate using Frame.io's formula: https://blog.frame.io/2017/03/06/calculate-video-bitrates/
        # This formula provides good quality while maintaining reasonable file sizes
        stream_duration = float(self.probe_box.duration) or float(video_stream.duration)
        if not stream_duration:
            pp.error("Could not calculate video duration")
            return input_path

        # Convert duration to minutes for bitrate calculation
        duration = stream_duration * 0.0166667

        stat = input_path.stat()
        pp.trace(f"File size: {stat}")
        file_size_megabytes = stat.st_size / 1000000

        # Calculate bitrates with a target of 50% of original size while maintaining quality
        current_bitrate = int(file_size_megabytes / (duration * 0.0075))
        target_bitrate = int(file_size_megabytes / (duration * 0.0075) / 2)
        # Allow 30% variance from target bitrate to handle complex scenes
        min_bitrate = int(current_bitrate * 0.7)
        max_bitrate = int(current_bitrate * 1.3)

        command: list[str] = ["-map", "0", "-c:v", "libx265"]
        command.extend(
            [
                "-b:v",
                f"{target_bitrate}k",
                "-minrate",
                f"{min_bitrate}k",
                "-maxrate",
                f"{max_bitrate}k",
                "-bufsize",
                f"{current_bitrate}k",
            ],
        )

        # Preserve original audio and subtitle streams to maintain quality
        command.extend(["-c:a", "copy", "-c:s", "copy"])

        return self._run_ffmpeg(command, title="Convert to H.265", step="h265")

    def convert_to_vp9(self) -> Path:
        """Convert the video to the VP9 codec format.

        Verify if conversion is required and proceed with it using ffmpeg. This method specifically targets the VP9 video codec. Return the path to the converted video or the original video if conversion is not necessary.

        Returns:
            Path: Path to the converted or original video file.
        """
        input_path = self.temp_file.latest_temp_path()

        video_stream = next(
            stream
            for stream in self.probe_box.streams
            if stream.codec_type == CodecTypes.VIDEO
            and stream.codec_name.lower() not in EXCLUDED_VIDEO_CODECS
        )

        if not video_stream:
            pp.error("No video stream found")
            return input_path

        # Skip re-encoding if already in modern codec unless forced
        if not settings.force and video_stream.codec_name.lower() in H265_CODECS:
            pp.warning(
                "VP9 ENCODE: Video already H.265 or VP9. Run with `--force` to re-encode. Skipping",
            )
            return input_path

        if Path(settings.out_path).suffix != ".webm":
            pp.info(
                f"Converting to VP9, setting output to `{settings.out_path.with_suffix('.webm').name}`"
            )
            settings.out_path = settings.out_path.with_suffix(".webm")

        # Use constant quality encoding (CRF) instead of bitrate for better quality control
        command: list[str] = [
            "-map",
            "0",
            "-c:v",
            "libvpx-vp9",
            "-b:v",
            "0",  # Disable fixed bitrate mode
            "-crf",
            "30",  # Higher CRF = lower quality but smaller file size
            "-c:a",
            "libvorbis",  # VP9 typically uses Vorbis audio codec
            "-dn",  # Disable data streams
            "-map_chapters",
            "-1",  # Remove chapters as they may cause issues in WebM
        ]

        command.extend(["-c:s", "copy"])

        return self._run_ffmpeg(command, title="Convert to vp9", suffix=".webm", step="vp9")

    def process_streams(self) -> Path:
        """Process the video file according to specified audio and subtitle preferences.

        Execute the necessary steps to process the video file, including managing audio and subtitle streams.  Keep or discard audio streams based on specified languages, commentary preferences, and downmix settings. Similarly, filter subtitle streams based on language preferences and criteria such as keeping commentary or local subtitles. Perform the processing using ffmpeg and return the path to the processed video file.

        Returns:
            Path: Path to the processed video file.
        """
        video_map_command = self._process_video()
        audio_map_command, downmix_command = self._process_audio()
        subtitle_map_command = self._process_subtitles()

        title_flags = []

        if audio_map_command:
            title_flags.append("drop original audio") if settings.drop_original_audio else None
            title_flags.append("keep commentary") if settings.keep_commentary else None
            title_flags.append("downmix to stereo") if settings.downmix_stereo else None

        if subtitle_map_command:
            title_flags.append(
                "keep subtitles",
            ) if settings.keep_all_subtitles else title_flags.append("drop unwanted subtitles")
            title_flags.append("keep local subtitles") if settings.keep_local_subtitles else None
            title_flags.append("drop local subtitles") if settings.drop_local_subs else None

        title = f"Process file ({', '.join(title_flags)})" if title_flags else "Process file"

        return self._run_ffmpeg(
            video_map_command
            + audio_map_command
            + subtitle_map_command
            + ["-c", "copy"]
            + downmix_command,
            title=title,
            step="process",
        )

    def reorder_streams(self) -> Path:
        """Reorder the media streams within the video file.

        Arrange the streams in the video file so that video streams appear first, followed by audio streams, and then subtitle streams. Exclude certain types of video streams like 'mjpeg' and 'png'.

        Returns:
            Path: Path to the video file with reordered streams.

        Raises:
            cappa.Exit: If no video or audio streams are found in the video file.
        """
        if not self.video_streams:
            pp.error("No video streams found")
            raise cappa.Exit(code=1)
        if not self.audio_streams:
            pp.error("No audio streams found")
            raise cappa.Exit(code=1)

        # Skip reordering if streams are already in the desired order (video->audio->subtitles)
        if not self._need_stream_reorder():
            pp.info(f"{SYMBOL_CHECK} No streams to reorder")
            return self.temp_file.latest_temp_path()

        # Use -c copy to avoid re-encoding when reordering streams
        initial_command = ["-c", "copy"]

        # Flatten stream lists into ffmpeg mapping commands while preserving desired order
        command = initial_command + [
            item
            for stream_list in [self.video_streams, self.audio_streams, self.subtitle_streams]
            for stream in stream_list
            for item in ["-map", f"0:{stream.index}"]
        ]

        return self._run_ffmpeg(command, title="Reorder streams", step="reorder")

    def video_to_1080p(self) -> Path:
        """Convert video resolution to 1080p.

        Scale video dimensions to 1920x1080 while maintaining aspect ratio. Only converts videos larger than 1080p unless forced.

        Returns:
            Path: Path to the converted video file, or original path if no conversion needed
        """
        input_path = self.temp_file.latest_temp_path()

        # Find first valid video stream, excluding thumbnail/image streams
        video_stream = next(
            stream
            for stream in self.probe_box.streams
            if stream.codec_type == CodecTypes.VIDEO
            and stream.codec_type.value not in EXCLUDED_VIDEO_CODECS
        )

        if not video_stream:
            pp.error("No video stream found")
            return input_path

        # Skip downscaling if video is already 1080p or smaller, unless forced
        if not settings.force and getattr(video_stream, "width", 0) <= 1920:  # noqa: PLR2004
            pp.info(f"{SYMBOL_CHECK} No convert to 1080p needed")
            return input_path

        # Use -2 for height to maintain aspect ratio while ensuring even dimensions for compatibility
        command: list[str] = [
            "-filter:v",
            "scale=width=1920:height=-2",
            "-c:a",
            "copy",
            "-c:s",
            "copy",
        ]

        return self._run_ffmpeg(command, title="Convert to 1080p", step="1080p")

    def ffprobe_json(self) -> dict:
        """Run ffprobe on the video file and return the JSON response.

        Returns:
            dict: A dictionary containing the ffprobe output with information about the video file's streams, format, and metadata.
        """
        return run_ffprobe(self.path)



"""Test the vidcleaner command line interface."""

import cappa
import pytest

from vid_cleaner.vidcleaner import VidCleaner, config_subcommand


@pytest.mark.parametrize(
    ("subcommand"),
    [("inspect"), ("clip"), ("clean"), ("cache")],
)
def test_vidcleaner_cli_help(clean_stdout, subcommand: str) -> None:
    """Verify help text displays for each subcommand."""
    # Given: Command line arguments requesting help
    args = [subcommand, "--help"] if subcommand else ["--help"]

    # When: Invoking CLI with help flag
    with pytest.raises(cappa.Exit):
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Help output contains expected information
    output = clean_stdout()
    assert "Usage: vidcleaner" in output
    assert "--help" in output
    assert " [-v]" in output



"""Test the vidcleaner inspect subcommand."""

import re
from pathlib import Path

import cappa
import pytest

from vid_cleaner import settings
from vid_cleaner.vidcleaner import VidCleaner, config_subcommand


def test_inspect_table(tmp_path, clean_stdout, debug, mock_video_path, mock_ffprobe_box, mocker):
    """Verify inspect command displays video information in table format."""
    # Given: Mock video file and ffprobe data
    args = ["inspect", str(mock_video_path)]
    settings.update({"cache_dir": Path(tmp_path), "langs_to_keep": ["en"]})
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )

    # When: Running inspect command
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Output contains expected table data
    output = clean_stdout()
    assert exc_info.value.code == 0
    assert re.search(r"0 │ video +│ h264", output)
    assert re.search(r"9 │ video +│ mjpeg", output)
    assert re.search(r"eng +│ 8 +│ 7.1", output)
    assert re.search(r"1920 +│ 1080 +│ Test", output)


def test_inspect_json(tmp_path, clean_stdout, debug, mock_video_path, mock_ffprobe, mocker):
    """Verify inspect command displays video information in JSON format."""
    # Given: Mock video file and ffprobe data
    args = ["inspect", "--json", str(mock_video_path)]
    settings.update({"cache_dir": Path(tmp_path), "langs_to_keep": ["en"]})
    mocker.patch(
        "vid_cleaner.models.video_file.run_ffprobe",
        return_value=mock_ffprobe("reference.json"),
    )

    # When: Running inspect command with JSON flag
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Output contains expected JSON data
    output = clean_stdout()
    assert exc_info.value.code == 0
    assert "'bit_rate': '26192239'," in output
    assert "'channel_layout': '7.1'," in output
    assert "'codec_name': 'hdmv_pgs_subtitle'," in output



"""Test the vidcleaner clean subcommand."""

from pathlib import Path

import cappa
import pytest
from iso639 import Lang

from vid_cleaner.vidcleaner import VidCleaner, config_subcommand

from vid_cleaner.models.video_file import VideoFile  # isort: skip
from vid_cleaner.controllers import TempFile  # isort: skip
from vid_cleaner import settings


@pytest.fixture(autouse=True)
def set_default_settings(tmp_path, mocker):
    """Set default settings for tests."""
    cache_dir = Path(tmp_path) / "cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    settings.update(
        {
            "cache_dir": cache_dir,
            "langs_to_keep": ["en"],
            "downmix_stereo": False,
            "keep_local_subtitles": False,
            "keep_commentary": False,
            "drop_local_subs": False,
            "keep_all_subtitles": False,
            "drop_original_audio": False,
            "save_each_step": False,
        }
    )


def test_fail_on_flag_conflict(debug, tmp_path, clean_stdout, mock_video_path):
    """Verify clean command fails when incompatible flags are used."""
    # Given: Conflicting codec conversion flags
    args = ["clean", "-vv", "--h265", "--vp9", str(mock_video_path)]
    settings.update({"cache_dir": Path(tmp_path), "langs_to_keep": ["en"]})

    # When: Running clean command with conflicting flags
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    # Then: Error message is displayed
    output = clean_stdout()

    assert exc_info.value.code == 1
    assert "Cannot convert to both H265 and VP9" in output


@pytest.mark.parametrize(
    ("args", "command_expected", "process_output"),
    [
        pytest.param(
            [],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4",
            "✔ Process file",
            id="Defaults (only keep local audio,no commentary)",
        ),
        pytest.param(
            ["--downmix"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4",
            "✔ Process file (downmix to stereo, drop unwanted subtitles)",
            id="Don't convert audio to stereo when stereo exists",
        ),
        pytest.param(
            ["--keep-commentary"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4 -map 0:5",
            "✔ Process file (keep commentary, drop unwanted subtitles)",
            id="Keep commentary",
        ),
        pytest.param(
            ["--drop-original"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4",
            "✔ Process file (drop original audio, drop unwanted subtitles)",
            id="Keep local language from config even when dropped",
        ),
        pytest.param(
            ["--langs", "fr,es"],
            "-map 0:0 -map 0:3 -map 0:8",
            "✔ Process file (drop unwanted subtitles)",
            id="Keep specified languages",
        ),
        pytest.param(
            ["--keep-subs"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4 -map 0:6 -map 0:7 -map 0:8",
            "✔ Process file (keep subtitles)",
            id="Keep all subtitles",
        ),
        pytest.param(
            ["--keep-local-subs"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4 -map 0:6",
            "✔ Process file (drop unwanted subtitles, keep local subtitles)",
            id="Keep local subtitles",
        ),
    ],
)
def test_stream_processing(
    debug,
    mocker,
    mock_ffprobe_box,
    mock_ffmpeg,
    clean_stdout,
    mock_video_path,
    args,
    command_expected,
    process_output,
) -> None:
    """Verify clean command processes video streams according to specified options."""
    # Given: Mock video file and processing options
    args = ["clean", "-vv", *args, str(mock_video_path)]

    # And: Mocked external dependencies
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )
    mocker.patch("vid_cleaner.cli.clean_video.copy_file", return_value="cleaned_video.mkv")
    mocker.patch.object(VideoFile, "_find_original_language", return_value=[Lang("en")])

    # When: Running clean command
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # Then: FFmpeg is called with correct stream mapping
    mock_ffmpeg.assert_called_once()
    args, _ = mock_ffmpeg.call_args
    command = " ".join(args[0])
    assert command_expected in command

    # And: Success messages are displayed
    assert exc_info.value.code == 0
    assert "✔ No streams to reorder" in output
    assert process_output in output
    assert "✅ Success: cleaned_video.mkv" in output


@pytest.mark.parametrize(
    ("args", "command_expected", "process_output"),
    [
        pytest.param(
            [],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4 -map 0:6",
            "✔ Process file (drop unwanted subtitles)",
            id="Defaults keep local and original audio, local subs",
        ),
        pytest.param(
            ["--drop-original"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4 -map 0:6",
            "✔ Process file (drop original audio, drop unwanted subtitles)",
            id="Drop original audio (keeps local audio)",
        ),
        pytest.param(
            ["--drop-local-subs"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4",
            "✔ Process file",
            id="Drop local subs",
        ),
    ],
)
def test_clean_video_foreign_language(
    mocker,
    mock_video_path,
    clean_stdout,
    tmp_path,
    mock_ffprobe_box,
    mock_ffmpeg,
    debug,
    args,
    command_expected,
    process_output,
):
    """Verify that video cleaning correctly processes foreign language videos."""
    args = ["clean", "-vv", *args, str(mock_video_path)]

    # And: Mock external dependencies
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )
    mocker.patch("vid_cleaner.cli.clean_video.copy_file", return_value="cleaned_video.mkv")
    mocker.patch.object(VideoFile, "_find_original_language", return_value=[Lang("fr")])

    # When: Processing the video file
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # THEN verify the ffmpeg command contains expected stream mappings
    mock_ffmpeg.assert_called_once()
    args, _ = mock_ffmpeg.call_args
    command = " ".join(args[0])

    # AND verify the command output indicates successful processing
    assert exc_info.value.code == 0
    assert command_expected in command
    assert "✔ No streams to reorder" in output
    assert process_output in output
    assert "✅ Success: cleaned_video.mkv" in output


@pytest.mark.parametrize(
    ("args", "command_expected", "process_output"),
    [
        pytest.param(
            [],
            "-map 0:0 -map 0:1 -map 0:2",
            "✔ Process file",
            id="Defaults, drops commentary",
        ),
        pytest.param(
            ["--downmix"],
            "-map 0:1 -map 0:2 -c copy -map 0:2 -c:a:0 aac -ac:a:0 2 -b:a:0 256k -filter:a:0",
            "✔ Process file (downmix to stereo)",
            id="Defaults",
        ),
    ],
)
def test_clean_video_downmix(
    mocker,
    mock_ffprobe_box,
    mock_video_path,
    clean_stdout,
    tmp_path,
    mock_ffmpeg,
    debug,
    args,
    command_expected,
    process_output,
):
    """Verify that videos without stereo audio are correctly downmixed."""
    args = ["clean", "-vv", *args, str(mock_video_path)]

    # And: Mock external dependencies
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("no_stereo.json"),
    )
    mocker.patch("vid_cleaner.cli.clean_video.copy_file", return_value="cleaned_video.mkv")
    mocker.patch.object(VideoFile, "_find_original_language", return_value=[Lang("en")])

    # When: Processing the video file
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # Then: FFmpeg is called with correct stream mapping
    mock_ffmpeg.assert_called_once()
    args, _ = mock_ffmpeg.call_args
    command = " ".join(args[0])
    assert command_expected in command

    # And: Success messages are displayed
    assert exc_info.value.code == 0
    assert "✔ No streams to reorder" in output
    assert process_output in output
    assert "✅ Success: cleaned_video.mkv" in output


@pytest.mark.parametrize(
    ("args", "first_command_expected", "second_command_expected", "process_output"),
    [
        pytest.param(
            [],
            "-c copy -map 0:2 -map 0:1 -map 0:3 -map 0:0",
            "-map 0:2 -map 0:1 -map 0:3",
            "✔ Process file",
            id="Defaults, reorder streams, then process streams",
        ),
    ],
)
def test_clean_reorganize_streams(
    mocker,
    mock_ffprobe_box,
    mock_video_path,
    tmp_path,
    clean_stdout,
    mock_ffmpeg,
    debug,
    args,
    first_command_expected,
    second_command_expected,
    process_output,
):
    """Verify that videos with incorrect stream order are properly reorganized."""
    args = ["clean", "-vv", *args, str(mock_video_path)]

    # And: Mock external dependencies
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("wrong_order.json"),
    )
    mocker.patch("vid_cleaner.cli.clean_video.copy_file", return_value="cleaned_video.mkv")
    mocker.patch.object(VideoFile, "_find_original_language", return_value=[Lang("en")])

    # When: Processing the video file
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # THEN verify ffmpeg is called twice - once to reorder streams and once to process them
    assert mock_ffmpeg.call_count == 2

    # AND verify the first ffmpeg command contains expected stream reordering
    first_command = " ".join(mock_ffmpeg.mock_calls[0].args[0])
    assert first_command_expected in first_command

    # AND verify the second ffmpeg command contains expected stream processing
    second_command = " ".join(mock_ffmpeg.mock_calls[2].args[0])
    assert second_command_expected in second_command

    # AND verify the command output indicates successful processing
    assert exc_info.value.code == 0
    assert "✔ Reorder streams" in output
    assert process_output in output
    assert "✅ Success: cleaned_video.mkv" in output


@pytest.mark.parametrize(
    ("args", "first_command_expected", "second_command_expected", "process_output"),
    [
        pytest.param(
            ["--h265"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4",
            "-map 0 -c:v libx265 -b:v 0k -minrate 0k -maxrate 0k -bufsize 0k -c:a copy -c:s copy",
            "✔ Process file",
            id="Convert to h265",
        ),
        pytest.param(
            ["--vp9"],
            "-map 0:0 -map 0:1 -map 0:2 -map 0:4",
            "-map 0 -c:v libvpx-vp9 -b:v 0 -crf 30 -c:a libvorbis -dn -map_chapters -1 -c:s copy",
            "✔ Process file",
            id="Convert to vp9",
        ),
    ],
)
def test_convert_video(
    mocker,
    mock_ffprobe_box,
    mock_video_path,
    tmp_path,
    mock_ffmpeg,
    clean_stdout,
    debug,
    args,
    first_command_expected,
    second_command_expected,
    process_output,
):
    """Verify video stream conversion with different codecs."""
    args = ["clean", "-vv", *args, str(mock_video_path)]
    # And: Mock external dependencies
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )
    mocker.patch("vid_cleaner.cli.clean_video.copy_file", return_value="cleaned_video.mkv")
    mocker.patch.object(VideoFile, "_find_original_language", return_value=[Lang("en")])
    mocker.patch.object(TempFile, "new_tmp_path", return_value=(mock_video_path))
    mocker.patch.object(TempFile, "latest_temp_path", return_value=(mock_video_path))

    # When: Processing the video file
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # THEN verify ffmpeg executes two passes
    assert mock_ffmpeg.call_count == 2

    # AND verify stream mapping command is correct
    first_command = " ".join(mock_ffmpeg.mock_calls[0].args[0])
    assert first_command_expected in first_command

    # AND verify codec conversion command is correct
    second_command = " ".join(mock_ffmpeg.mock_calls[2].args[0])
    assert second_command_expected in second_command

    # AND verify successful completion with expected output messages
    assert exc_info.value.code == 0
    assert "✔ No streams to reorder" in output
    assert process_output in output
    assert "✅ Success: cleaned_video.mkv" in output

    if "--vp9" in args:
        assert "Converting to VP9, setting output to test_video.webm" in output


def test_save_each_step(
    mocker,
    mock_ffprobe_box,
    mock_video_path,
    mock_ffmpeg,
    clean_stdout,
    tmp_path,
    debug,
):
    """Verify video stream conversion with different codecs."""
    # GIVEN a second video file used to mock the output of the first video file
    mock_video_path_1 = Path(tmp_path / "test_video_1.mp4")
    mock_video_path_1.touch()  # Create a dummy file

    args = ["clean", "-vv", "--save-each", "--h265", "--downmix", str(mock_video_path)]
    # And: Mock external dependencies
    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )
    mocker.patch(
        "vid_cleaner.cli.clean_video.copy_file",
        side_effect=[mock_video_path_1, "cleaned_video.mkv"],
    )
    mocker.patch.object(VideoFile, "_find_original_language", return_value=[Lang("en")])
    mocker.patch.object(TempFile, "new_tmp_path", return_value=(mock_video_path))
    mocker.patch.object(TempFile, "latest_temp_path", return_value=(mock_video_path))

    # When: Processing the video file
    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # debug(output, "output")

    # THEN verify ffmpeg executes two passes
    assert mock_ffmpeg.call_count == 2

    # AND verify stream mapping command is correct
    first_command = " ".join(mock_ffmpeg.mock_calls[0].args[0])
    assert "-map 0:0 -map 0:1 -map 0:2 -map 0:4" in first_command

    # AND verify codec conversion command is correct
    second_command = " ".join(mock_ffmpeg.mock_calls[2].args[0])
    assert (
        "-map 0 -c:v libx265 -b:v 0k -minrate 0k -maxrate 0k -bufsize 0k -c:a copy -c:s copy"
        in second_command
    )

    # AND verify successful completion with expected output messages
    assert exc_info.value.code == 0
    assert "✔ No streams to reorder" in output
    assert "✔ Process file" in output
    assert "✅ Success: cleaned_video.mkv" in output
    assert "✔ Process file (downmix to stereo, drop unwanted subtitles)" in output
    assert "✅ Success: …/test_video_1.mp4" in output
    assert "✔ Convert to H.265" in output
    assert "✅ Success: cleaned_video.mkv" in output



"""Clean subcommand."""

from pathlib import Path

import cappa
from nclutils import copy_file, pp

from vid_cleaner import settings
from vid_cleaner.utils import coerce_video_files
from vid_cleaner.vidcleaner import CleanCommand

from vid_cleaner.models.video_file import VideoFile  # isort: skip


def main(clean_cmd: CleanCommand) -> None:
    """Process video files according to specified cleaning options.

    Apply video processing operations like stream reordering, audio/subtitle filtering, and format conversion based on command line arguments.

    Args:
        cmd (VidCleaner): Global command options and configuration
        clean_cmd (CleanCommand): Clean-specific command options

    Raises:
        cappa.Exit: If incompatible options are specified (e.g., both H265 and VP9)
    """
    if settings.h265 and settings.vp9:
        pp.error("Cannot convert to both H265 and VP9")
        raise cappa.Exit(code=1)

    for video in coerce_video_files(clean_cmd.files):
        settings.out_path = settings.out_path or video.path

        pp.info(f"⇨ {video.path.name}")
        video.reorder_streams()
        video.process_streams()

        if not settings.dryrun and settings.save_each_step:
            out_file = copy_file(
                src=video.temp_file.latest_temp_path(),
                dst=settings.out_path,
                keep_backup=not settings.overwrite,
                with_progress=True,
                transient=True,
            )
            pp.success(f"{out_file}")
            video.temp_file.clean_up()

            video = VideoFile(Path(out_file))  # noqa: PLW2901

        if settings.video_1080:
            video.video_to_1080p()

            if not settings.dryrun and settings.save_each_step:
                out_file = copy_file(
                    src=video.temp_file.latest_temp_path(),
                    dst=settings.out_path,
                    keep_backup=not settings.overwrite,
                    with_progress=True,
                    transient=True,
                )
                pp.success(f"{out_file}")
                video.temp_file.clean_up()

                video = VideoFile(Path(out_file))  # noqa: PLW2901

        if settings.h265:
            video.convert_to_h265()

        if settings.vp9:
            video.convert_to_vp9()

        if not settings.dryrun:
            out_file = copy_file(
                src=video.temp_file.latest_temp_path(),
                dst=settings.out_path,
                keep_backup=not settings.overwrite,
                with_progress=True,
                transient=True,
            )
            video.temp_file.clean_up()

            if settings.overwrite and out_file != video.path:
                pp.debug(f"Delete: {video.path}")
                video.path.unlink()

            pp.success(f"{out_file}")

    raise cappa.Exit(code=0)



# type: ignore
"""Test the search command."""

from pathlib import Path

import cappa
import pytest

from vid_cleaner import settings
from vid_cleaner.vidcleaner import VidCleaner, config_subcommand


@pytest.fixture(autouse=True)
def set_default_settings(tmp_path, mocker, mock_ffprobe_box):
    """Set default settings for tests."""
    cache_dir = Path(tmp_path) / "cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    settings.update(
        {
            "cache_dir": cache_dir,
            "langs_to_keep": ["en"],
            "downmix_stereo": False,
            "keep_local_subtitles": False,
            "keep_commentary": False,
            "drop_local_subs": False,
            "keep_all_subtitles": False,
            "drop_original_audio": False,
            "save_each_step": False,
        }
    )


def test_search_no_video_files(tmp_path, clean_stdout, mocker, debug):
    """Test that the search command returns no results when there are no video files."""
    # Given: A directory with no video files
    directory = Path(tmp_path) / "no_videos"
    directory.mkdir(parents=True, exist_ok=True)

    # When: Running the search command
    args = ["search", str(directory)]

    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # Then: The command should return no results
    # debug(output, "output")
    assert exc_info.value.code == 0
    assert "No video files found" in output


def test_search_with_results(
    tmp_path,
    clean_stdout,
    mocker,
    mock_video_path,
    # mock_video_file,
    debug,
    mock_ffprobe_box,
    mock_ffprobe,
):
    """Test that the search command returns results when there are video files."""
    # When: Running the search command
    args = ["search", str(mock_video_path.parent), "--filters", "h264,reorder"]

    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )

    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # Then: The command should return results
    # debug(output, "output")

    assert exc_info.value.code == 0
    assert "Found 1 video files in 1/1 directories" in output
    assert "1 │ test_video.mp4 │ h264" in output


def test_search_with_no_results(
    tmp_path,
    clean_stdout,
    mocker,
    mock_video_path,
    # mock_video_file,
    debug,
    mock_ffprobe_box,
    mock_ffprobe,
):
    """Test that the search command returns results when there are video files."""
    # When: Running the search command
    args = ["search", str(mock_video_path.parent), "--filters", "reorder"]

    mocker.patch(
        "vid_cleaner.models.video_file.get_probe_as_box",
        return_value=mock_ffprobe_box("reference.json"),
    )

    with pytest.raises(cappa.Exit) as exc_info:
        cappa.invoke(obj=VidCleaner, argv=args, deps=[config_subcommand])

    output = clean_stdout()
    # Then: The command should return results
    # debug(output, "output")

    assert exc_info.value.code == 1
    assert "Found 1 video files in 1/1 directories" in output
    assert "No video files found matching 'reorder'" in output
    assert "Matching filters" not in output



"""Cache subcommand."""

import shutil

import cappa
from nclutils import console, directory_tree, pp

from vid_cleaner import settings
from vid_cleaner.vidcleaner import CacheCommand


def main(cache_cmd: CacheCommand) -> None:
    """Manage the cache directory for vidcleaner.

    Clear or display the contents of the cache directory used by vidcleaner. When displaying contents, show a tree view of all files including hidden ones.

    Args:
        cmd (VidCleaner): The main command instance containing global options
        cache_cmd (CacheCommand): The cache subcommand instance with cache-specific options

    Raises:
        cappa.Exit: If the cache directory is not found or if the cache is cleared
    """
    if not settings.CACHE_DIR.exists():
        pp.info(f"Cache directory not found: `{settings.CACHE_DIR}`")
        raise cappa.Exit(code=0)

    if cache_cmd.clear:
        shutil.rmtree(settings.CACHE_DIR)
        settings.CACHE_DIR.mkdir(parents=True, exist_ok=True)
        pp.success(f"Cache cleared: `{settings.CACHE_DIR}`")
        raise cappa.Exit(code=0)

    if len(list(settings.CACHE_DIR.iterdir())) > 0:
        tree = directory_tree(settings.CACHE_DIR, show_hidden=True)
        console.print(tree)
    else:
        pp.info(f"Cache directory is empty: `{settings.CACHE_DIR}`")

    raise cappa.Exit(code=0)
